import dataclasses
import typing

from . import contributions as _contributions
from . import enums as _enums
from . import math as _math
from . import string as _string

_EMPTY_CHORD = "<>"


# TODO: move to indicators.py
# TODO: frozen=True
@dataclasses.dataclass(eq=False, slots=True, unsafe_hash=True)
class Dynamic:
    r"""
    Dynamic.

    ..  container:: example

        Initializes from dynamic name:

        >>> voice = abjad.Voice("c'8 d'8 e'8 f'8")
        >>> dynamic = abjad.Dynamic("f")
        >>> abjad.attach(dynamic, voice[0])

        ..  docs::

            >>> string = abjad.lilypond(voice)
            >>> print(string)
            \new Voice
            {
                c'8
                \f
                d'8
                e'8
                f'8
            }

        >>> abjad.show(voice) # doctest: +SKIP

    ..  container:: example

        Initializes from other dynamic:

        >>> dynamic_1 = abjad.Dynamic("f")
        >>> dynamic_2 = abjad.Dynamic(dynamic_1)

        >>> dynamic_1
        Dynamic(name='f', command=None, format_hairpin_stop=False, hide=False, leak=False, name_is_textual=False, ordinal=2)

        >>> dynamic_2
        Dynamic(name='f', command=None, format_hairpin_stop=False, hide=False, leak=False, name_is_textual=False, ordinal=2)

    ..  container:: example

        Initializes niente:

        >>> abjad.Dynamic("niente")
        Dynamic(name='niente', command=None, format_hairpin_stop=False, hide=False, leak=False, name_is_textual=True, ordinal=NegativeInfinity())

    ..  container:: example

        Simultaneous dynamics in a single staff:

        >>> voice_1 = abjad.Voice("e'8 g'8 f'8 a'8")
        >>> abjad.attach(abjad.Dynamic('f'), voice_1[0], context='Voice')
        >>> literal = abjad.LilyPondLiteral(r"\voiceOne", "opening")
        >>> abjad.attach(literal, voice_1)
        >>> abjad.override(voice_1).DynamicLineSpanner.direction = abjad.UP
        >>> voice_2 = abjad.Voice("c'2")
        >>> literal = abjad.LilyPondLiteral(r"\voiceTwo", "opening")
        >>> abjad.attach(literal, voice_2)
        >>> abjad.attach(abjad.Dynamic("mf"), voice_2[0], context="Voice")
        >>> staff = abjad.Staff([voice_1, voice_2], simultaneous=True)
        >>> abjad.show(staff) # doctest: +SKIP

        ..  docs::

            >>> string = abjad.lilypond(staff)
            >>> print(string)
            \new Staff
            <<
                \new Voice
                \with
                {
                    \override DynamicLineSpanner.direction = #up
                }
                {
                    \voiceOne
                    e'8
                    \f
                    g'8
                    f'8
                    a'8
                }
                \new Voice
                {
                    \voiceTwo
                    c'2
                    \mf
                }
            >>

        >>> for leaf in abjad.select.leaves(staff):
        ...     dynamic = abjad.get.effective(leaf, abjad.Dynamic)
        ...     print(f"{leaf!r}:")
        ...     print(f"    {dynamic!r}")
        Note("e'8"):
            Dynamic(name='f', command=None, format_hairpin_stop=False, hide=False, leak=False, name_is_textual=False, ordinal=2)
        Note("g'8"):
            Dynamic(name='f', command=None, format_hairpin_stop=False, hide=False, leak=False, name_is_textual=False, ordinal=2)
        Note("f'8"):
            Dynamic(name='f', command=None, format_hairpin_stop=False, hide=False, leak=False, name_is_textual=False, ordinal=2)
        Note("a'8"):
            Dynamic(name='f', command=None, format_hairpin_stop=False, hide=False, leak=False, name_is_textual=False, ordinal=2)
        Note("c'2"):
            Dynamic(name='mf', command=None, format_hairpin_stop=False, hide=False, leak=False, name_is_textual=False, ordinal=1)

    ..  container:: example exception

        Errors on nondynamic input:

        >>> abjad.Dynamic("text")
        Traceback (most recent call last):
            ...
        Exception: letter 't' (in 'text') is not a dynamic.

    ..  container:: example

        REGRESSION. Duplicate dynamics raise exception on attach:

        >>> staff = abjad.Staff("c'4 d' e' f'")
        >>> dynamic = abjad.Dynamic("p")
        >>> abjad.attach(dynamic, staff[0])
        >>> dynamic = abjad.Dynamic("p")
        >>> abjad.attach(dynamic, staff[0], check_duplicate_indicator=True)
        Traceback (most recent call last):
            ...
        abjad.exceptions.PersistentIndicatorError:
        <BLANKLINE>
        Can not attach ...
        <BLANKLINE>
        Wrapper(...)
        <BLANKLINE>
            ... to Note("c'4") in None because ...
        <BLANKLINE>
        Wrapper(...)
        <BLANKLINE>
            ... is already attached to the same leaf.

    ..  container:: example

        Tweaks:

        >>> note = abjad.Note("c'4")
        >>> dynamic = abjad.Dynamic("f")
        >>> bundle = abjad.bundle(dynamic, r"- \tweak color #blue")
        >>> abjad.attach(bundle, note)
        >>> abjad.show(note) # doctest: +SKIP

        ..  docs::

            >>> string = abjad.lilypond(note)
            >>> print(string)
            c'4
            - \tweak color #blue
            \f

    ..  container:: example

        Use ``command`` like this:

        >>> abjad.Dynamic("f", command=r"\sub_f").command
        '\\sub_f'

        Use to override LilyPond output when a custom dynamic has been defined in an
        external stylesheet. (In the example above, ``\sub_f`` is a nonstandard LilyPond
        dynamic. LilyPond will interpret the output above only when the command
        ``\sub_f`` is defined somewhere in an external stylesheet.)


    ..  container:: example

        Direction:

        With ``direction`` unset:

        >>> staff = abjad.Staff("c'2 c''2")
        >>> abjad.attach(abjad.Dynamic("p"), staff[0])
        >>> abjad.attach(abjad.Dynamic("f"), staff[1])
        >>> abjad.show(staff) # doctest: +SKIP

        ..  docs::

            >>> string = abjad.lilypond(staff)
            >>> print(string)
            \new Staff
            {
                c'2
                \p
                c''2
                \f
            }

        With ``direction=abjad.UP``:

        >>> staff = abjad.Staff("c'2 c''2")
        >>> abjad.attach(abjad.Dynamic("p"), staff[0], direction=abjad.UP)
        >>> abjad.attach(abjad.Dynamic("f"), staff[1], direction=abjad.UP)
        >>> abjad.show(staff) # doctest: +SKIP

        ..  docs::

            >>> string = abjad.lilypond(staff)
            >>> print(string)
            \new Staff
            {
                c'2
                ^ \p
                c''2
                ^ \f
            }

        With ``direction=abjad.DOWN``:

        >>> staff = abjad.Staff("c'2 c''2")
        >>> abjad.attach(abjad.Dynamic("p"), staff[0], direction=abjad.DOWN)
        >>> abjad.attach(abjad.Dynamic("f"), staff[1], direction=abjad.DOWN)
        >>> abjad.show(staff) # doctest: +SKIP

        ..  docs::

            >>> string = abjad.lilypond(staff)
            >>> print(string)
            \new Staff
            {
                c'2
                _ \p
                c''2
                _ \f
            }

    ..  container:: example

        REGRESSION. Effort dynamics default to down:

        >>> staff = abjad.Staff("c'2 c''2")
        >>> abjad.attach(abjad.Dynamic('"p"'), staff[0])
        >>> abjad.attach(abjad.Dynamic('"f"'), staff[1])
        >>> abjad.show(staff) # doctest: +SKIP

        ..  docs::

            >>> string = abjad.lilypond(staff)
            >>> print(string)
            \new Staff
            {
                c'2
                _ #(make-dynamic-script
                    (markup
                        #:whiteout
                        #:line (
                            #:general-align Y -2 #:normal-text #:larger "“"
                            #:hspace -0.1
                            #:dynamic "p"
                            #:hspace -0.25
                            #:general-align Y -2 #:normal-text #:larger "”"
                            )
                        )
                    )
                c''2
                _ #(make-dynamic-script
                    (markup
                        #:whiteout
                        #:line (
                            #:general-align Y -2 #:normal-text #:larger "“"
                            #:hspace -0.4
                            #:dynamic "f"
                            #:hspace -0.2
                            #:general-align Y -2 #:normal-text #:larger "”"
                            )
                        )
                    )
            }

        And may be overriden:

        >>> staff = abjad.Staff("c'2 c''2")
        >>> abjad.attach(abjad.Dynamic('"p"'), staff[0], direction=abjad.UP)
        >>> abjad.attach(abjad.Dynamic('"f"'), staff[1], direction=abjad.UP)
        >>> abjad.show(staff) # doctest: +SKIP

        ..  docs::

            >>> string = abjad.lilypond(staff)
            >>> print(string)
            \new Staff
            {
                c'2
                ^ #(make-dynamic-script
                    (markup
                        #:whiteout
                        #:line (
                            #:general-align Y -2 #:normal-text #:larger "“"
                            #:hspace -0.1
                            #:dynamic "p"
                            #:hspace -0.25
                            #:general-align Y -2 #:normal-text #:larger "”"
                            )
                        )
                    )
                c''2
                ^ #(make-dynamic-script
                    (markup
                        #:whiteout
                        #:line (
                            #:general-align Y -2 #:normal-text #:larger "“"
                            #:hspace -0.4
                            #:dynamic "f"
                            #:hspace -0.2
                            #:general-align Y -2 #:normal-text #:larger "”"
                            )
                        )
                    )
            }

    ..  container:: example

        Set ``hide=True`` when dynamic should not appear in output (but should still
        determine effective dynamic):

        >>> voice = abjad.Voice("c'4 d' e' f'")
        >>> abjad.attach(abjad.Dynamic("f"), voice[0])
        >>> abjad.attach(abjad.Dynamic("mf", hide=True), voice[2])
        >>> abjad.show(voice) # doctest: +SKIP

        >>> string = abjad.lilypond(voice)
        >>> print(string)
        \new Voice
        {
            c'4
            \f
            d'4
            e'4
            f'4
        }

        >>> for leaf in abjad.iterate.leaves(voice):
        ...     dynamic = abjad.get.effective(leaf, abjad.Dynamic)
        ...     print(f"{leaf!r}:")
        ...     print(f"    {dynamic!r}")
        Note("c'4"):
            Dynamic(name='f', command=None, format_hairpin_stop=False, hide=False, leak=False, name_is_textual=False, ordinal=2)
        Note("d'4"):
            Dynamic(name='f', command=None, format_hairpin_stop=False, hide=False, leak=False, name_is_textual=False, ordinal=2)
        Note("e'4"):
            Dynamic(name='mf', command=None, format_hairpin_stop=False, hide=True, leak=False, name_is_textual=False, ordinal=1)
        Note("f'4"):
            Dynamic(name='mf', command=None, format_hairpin_stop=False, hide=True, leak=False, name_is_textual=False, ordinal=1)

    ..  container:: example

        Set ``leak=True`` Is true to format LilyPond empty chord ``<>`` symbol:

        Without leaked stop dynamic:

        >>> staff = abjad.Staff("c'4 d' e' r")
        >>> start_dynamic = abjad.Dynamic("mf")
        >>> start_hairpin = abjad.StartHairpin(">")
        >>> stop_dynamic = abjad.Dynamic("pp")
        >>> abjad.attach(start_dynamic, staff[0])
        >>> abjad.attach(start_hairpin, staff[0])
        >>> abjad.attach(stop_dynamic, staff[-2])
        >>> abjad.override(staff).DynamicLineSpanner.staff_padding = 4
        >>> abjad.show(staff) # doctest: +SKIP

        ..  docs::

            >>> string = abjad.lilypond(staff)
            >>> print(string)
            \new Staff
            \with
            {
                \override DynamicLineSpanner.staff-padding = 4
            }
            {
                c'4
                \mf
                \>
                d'4
                e'4
                \pp
                r4
            }

        With leaked stop dynamic:

        >>> staff = abjad.Staff("c'4 d' e' r")
        >>> start_dynamic = abjad.Dynamic("mf")
        >>> start_hairpin = abjad.StartHairpin(">")
        >>> stop_dynamic = abjad.Dynamic("pp", leak=True)
        >>> abjad.attach(start_dynamic, staff[0])
        >>> abjad.attach(start_hairpin, staff[0])
        >>> abjad.attach(stop_dynamic, staff[-2])
        >>> abjad.override(staff).DynamicLineSpanner.staff_padding = 4
        >>> abjad.show(staff) # doctest: +SKIP

        ..  docs::

            >>> string = abjad.lilypond(staff)
            >>> print(string)
            \new Staff
            \with
            {
                \override DynamicLineSpanner.staff-padding = 4
            }
            {
                c'4
                \mf
                \>
                d'4
                e'4
                <>
                \pp
                r4
            }

    ..  container:: example

        Leaks format after spanners:

        >>> staff = abjad.Staff("c'8 [ d' e' ] f'")
        >>> start_dynamic = abjad.Dynamic("mf")
        >>> start_hairpin = abjad.StartHairpin(">")
        >>> stop_dynamic = abjad.Dynamic("pp", leak=True)
        >>> abjad.attach(start_dynamic, staff[0])
        >>> abjad.attach(start_hairpin, staff[0])
        >>> abjad.attach(stop_dynamic, staff[-2])
        >>> abjad.override(staff).DynamicLineSpanner.staff_padding = 4
        >>> abjad.show(staff) # doctest: +SKIP

        ..  docs::

            >>> string = abjad.lilypond(staff)
            >>> print(string)
            \new Staff
            \with
            {
                \override DynamicLineSpanner.staff-padding = 4
            }
            {
                c'8
                \mf
                [
                \>
                d'8
                e'8
                ]
                <>
                \pp
                f'8
            }

    ..  container:: example

        Leaked and nonleaked dynamic may be attached to the same leaf:

        >>> staff = abjad.Staff("c'4 d' e' f'")
        >>> abjad.attach(abjad.Dynamic("f"), staff[0])
        >>> abjad.attach(abjad.Dynamic("p", leak=True), staff[0])
        >>> abjad.show(staff) # doctest: +SKIP

        ..  docs::

            >>> string = abjad.lilypond(staff)
            >>> print(string)
            \new Staff
            {
                c'4
                \f
                <>
                \p
                d'4
                e'4
                f'4
            }

    ..  container:: example

        Leaks and tweaks on the same dynamic format correctly; LilyPond empty chord
        ``<>`` symbol appears before postevents:

        >>> staff = abjad.Staff("r4 d' e' f'")
        >>> dynamic = abjad.Dynamic("f", leak=True)
        >>> bundle = abjad.bundle(dynamic, r"- \tweak color #blue")
        >>> abjad.attach(bundle, staff[0])
        >>> abjad.show(staff) # doctest: +SKIP

        >>> string = abjad.lilypond(staff)
        >>> print(string)
        \new Staff
        {
            r4
            <>
            - \tweak color #blue
            \f
            d'4
            e'4
            f'4
        }

    ..  container:: example

        Leak survives copy:

        >>> import copy
        >>> dynamic = abjad.Dynamic("pp", leak=True)
        >>> copy.copy(dynamic)
        Dynamic(name='pp', command=None, format_hairpin_stop=False, hide=False, leak=True, name_is_textual=False, ordinal=-3)

    ..  container:: example

        Niente dynamics format like this:

        >>> voice = abjad.Voice("c'4 r r c'4")
        >>> abjad.attach(abjad.Dynamic("p"), voice[0])
        >>> abjad.attach(abjad.Dynamic("niente"), voice[1])
        >>> abjad.attach(abjad.Dynamic("p"), voice[3])
        >>> abjad.override(voice).DynamicLineSpanner.staff_padding = 4
        >>> abjad.show(voice) # doctest: +SKIP

        ..  docs::

            >>> string = abjad.lilypond(voice)
            >>> print(string)
            \new Voice
            \with
            {
                \override DynamicLineSpanner.staff-padding = 4
            }
            {
                c'4
                \p
                r4
                _ #(make-dynamic-script (markup #:whiteout #:normal-text #:italic "niente"))
                r4
                c'4
                \p
            }

    ..  container:: example

        Name-is-textual:

        >>> abjad.Dynamic("f").name_is_textual
        False

        >>> abjad.Dynamic("niente").name_is_textual
        True

        >>> dynamic = abjad.Dynamic("appena udibile", name_is_textual=True)
        >>> dynamic.name_is_textual
        True

        Textual dynamics format like this when initialized without an explicit command:

        >>> voice = abjad.Voice("c'4 d' e' f'")
        >>> dynamic = abjad.Dynamic("appena udibile", name_is_textual=True)
        >>> abjad.attach(dynamic, voice[0])
        >>> abjad.override(voice).DynamicLineSpanner.staff_padding = 4
        >>> abjad.override(voice).DynamicText.X_extent = "#'(0 . 0)"
        >>> abjad.override(voice).DynamicText.self_alignment_X = abjad.LEFT
        >>> abjad.show(voice) # doctest: +SKIP

        ..  docs::

            >>> string = abjad.lilypond(voice)
            >>> print(string)
            \new Voice
            \with
            {
                \override DynamicLineSpanner.staff-padding = 4
                \override DynamicText.X-extent = #'(0 . 0)
                \override DynamicText.self-alignment-X = #left
            }
            {
                c'4
                _ #(make-dynamic-script (markup #:whiteout #:normal-text #:italic "appena udibile"))
                d'4
                e'4
                f'4
            }

        Textual dynamics format like this when initialized with an explicit command:

        >>> voice = abjad.Voice("c'4 d' e' f'")
        >>> dynamic = abjad.Dynamic(
        ...     "appena udibile",
        ...     command=r"\appena_udibile",
        ...     name_is_textual=True,
        ... )
        >>> abjad.attach(dynamic, voice[0])

        Only LilyPond output is shown here because dynamic commands (like
        ``\appena_udibile`` shown here) are meant to be user-defined (and not included in
        Abjad):

        >>> string = abjad.lilypond(voice)
        >>> print(string)
        \new Voice
        {
            c'4
            \appena_udibile
            d'4
            e'4
            f'4
        }

        REGRESSION. Textual names work with replace:

        >>> import dataclasses
        >>> dynamic = abjad.Dynamic("niente")
        >>> dataclasses.replace(dynamic)
        Dynamic(name='niente', command=None, format_hairpin_stop=False, hide=False, leak=False, name_is_textual=True, ordinal=NegativeInfinity())

        >>> dynamic = abjad.Dynamic("appena udibile", name_is_textual=True)
        >>> dataclasses.replace(dynamic)
        Dynamic(name='appena udibile', command=None, format_hairpin_stop=False, hide=False, leak=False, name_is_textual=True, ordinal=None)

    ..  container:: example

        Ordinal value of a dynamic:

        >>> abjad.Dynamic("f").ordinal
        2

        >>> abjad.Dynamic("p").ordinal
        -2

        >>> abjad.Dynamic("niente").ordinal
        NegativeInfinity()

        >>> abjad.Dynamic('"f"').ordinal
        2

        >>> abjad.Dynamic('"p"').ordinal
        -2

        User-defined ordinals:

        >>> barely_audible = abjad.Dynamic(
        ...     "barely audible",
        ...     name_is_textual=True,
        ...     ordinal=-99,
        ... )
        >>> barely_audible.ordinal
        -99

        >>> extremely_loud = abjad.Dynamic(
        ...     "extremely loud",
        ...     name_is_textual=True,
        ...     ordinal=99,
        ... )
        >>> extremely_loud.ordinal
        99

        REGRESSION. Textual names without explicit ordinal return none:

        >>> dynamic = abjad.Dynamic("appena udibile", name_is_textual=True)
        >>> dynamic.ordinal is None
        True

    """

    name: typing.Union[str, "Dynamic"] = "f"
    command: str | None = None
    format_hairpin_stop: bool = False
    hide: bool = False
    leak: bool = dataclasses.field(default=False, compare=False)
    name_is_textual: bool = False
    ordinal: int | _math.Infinity | _math.NegativeInfinity | None = None

    context: typing.ClassVar[str] = "Voice"
    directed: typing.ClassVar[bool] = True
    parameter: typing.ClassVar[str] = "DYNAMIC"
    persistent: typing.ClassVar[bool] = True
    post_event: typing.ClassVar[bool] = True
    spanner_stop: typing.ClassVar[bool] = True

    def __post_init__(self):
        if self.name is not None:
            assert isinstance(self.name, str | Dynamic), repr(self.name)
        if isinstance(self.name, Dynamic):
            name_ = self.name.name
        elif isinstance(self.name, str):
            name_ = self.name
        if name_ == "niente":
            self.name_is_textual = True
        if not self.name_is_textual:
            for letter in name_.strip('"'):
                if letter not in self._lilypond_dynamic_alphabet:
                    message = f"letter {letter!r} (in {self.name!r}) is not a dynamic."
                    raise Exception(message)
        self.name = name_
        if self.command is not None:
            assert isinstance(self.command, str), repr(self.command)
            assert self.command.startswith("\\"), repr(self.command)
        assert isinstance(self.format_hairpin_stop, bool), repr(
            self.format_hairpin_stop
        )
        assert isinstance(self.hide, bool), repr(self.hide)
        assert isinstance(self.leak, bool), repr(self.leak)
        assert isinstance(self.name_is_textual, bool), repr(self.name_is_textual)
        if self.ordinal is not None:
            assert isinstance(
                self.ordinal, (int, _math.Infinity, _math.NegativeInfinity)
            )
        else:
            stripped_name = None
            if self.name:
                stripped_name = self.name.strip('"')
            if (
                stripped_name
                in self._composite_dynamic_name_to_steady_state_dynamic_name
            ):
                stripped_name = (
                    self._composite_dynamic_name_to_steady_state_dynamic_name[
                        stripped_name
                    ]
                )
            if stripped_name is None:
                ordinal = None
            else:
                assert isinstance(stripped_name, str), repr(stripped_name)
                ordinal_ = self._dynamic_name_to_dynamic_ordinal.get(stripped_name)
                prototype = (int, _math.Infinity, _math.NegativeInfinity, type(None))
                assert isinstance(ordinal_, prototype), repr(ordinal_)
                ordinal = ordinal_
            self.ordinal = ordinal

    _composite_dynamic_name_to_steady_state_dynamic_name: typing.ClassVar = {
        "fp": "p",
        "sf": "f",
        "sff": "ff",
        "sfp": "p",
        "sfpp": "pp",
        "sffp": "p",
        "sffpp": "pp",
        "sfz": "f",
        "sp": "p",
        "spp": "pp",
        "rfz": "f",
    }

    _dynamic_name_to_dynamic_ordinal: typing.ClassVar = {
        "ppppp": -6,
        "pppp": -5,
        "ppp": -4,
        "pp": -3,
        "p": -2,
        "niente": _math.NegativeInfinity(),
        "mp": -1,
        "mf": 1,
        "f": 2,
        "ff": 3,
        "fff": 4,
        "ffff": 5,
        "fffff": 6,
    }

    _dynamic_names: typing.ClassVar = (
        "ppppp",
        "pppp",
        "ppp",
        "pp",
        "p",
        "mp",
        "mf",
        "f",
        "ff",
        "fff",
        "ffff",
        "fffff",
        "fp",
        "sf",
        "sff",
        "sp",
        "spp",
        "sfz",
        "sffz",
        "sfffz",
        "sffp",
        "sffpp",
        "sfp",
        "sfpp",
        "rfz",
        "niente",
    )

    _dynamic_ordinal_to_dynamic_name: typing.ClassVar = {
        -6: "ppppp",
        -5: "pppp",
        -4: "ppp",
        -3: "pp",
        -2: "p",
        -1: "mp",
        _math.NegativeInfinity(): "niente",
        1: "mf",
        2: "f",
        3: "ff",
        4: "fff",
        5: "ffff",
        6: "fffff",
    }

    _site: typing.ClassVar[str] = "after"

    _lilypond_dynamic_commands: typing.ClassVar = [
        _ for _ in _dynamic_names if not _ == "niente"
    ]

    _lilypond_dynamic_alphabet: typing.ClassVar = "fmprsz"

    _to_width: typing.ClassVar = {
        '"f"': 2,
        '"mf"': 3.5,
        '"mp"': 3.5,
        '"p"': 2,
        "sfz": 2.5,
    }

    def __eq__(self, argument) -> bool:
        """
        Is true when ``argument`` equals dynamic.

        ..  container:: example

            >>> dynamic_1 = abjad.Dynamic("p")
            >>> dynamic_2 = abjad.Dynamic("p")
            >>> dynamic_3 = abjad.Dynamic("f")

            >>> dynamic_1 == dynamic_1
            True
            >>> dynamic_1 == dynamic_2
            True
            >>> dynamic_1 == dynamic_3
            False

            >>> dynamic_2 == dynamic_1
            True
            >>> dynamic_2 == dynamic_2
            True
            >>> dynamic_2 == dynamic_3
            False

            >>> dynamic_3 == dynamic_1
            False
            >>> dynamic_3 == dynamic_2
            False
            >>> dynamic_3 == dynamic_3
            True

        """
        if not isinstance(argument, type(self)):
            return False
        if self.name == argument.name and self.ordinal == argument.ordinal:
            return True
        return False

    def _attachment_test_all(self, component_expression):
        if not hasattr(component_expression, "written_duration"):
            strings = [f"Must be leaf (not {component_expression})."]
            return strings
        return True

    def _format_effort_dynamic(self, *, wrapper=None):
        name = self.name.strip('"')
        before = {"f": -0.4, "m": -0.1, "p": -0.1, "r": -0.1, "s": -0.3, "z": -0.2}[
            name[0]
        ]
        after = {"f": -0.2, "m": -0.1, "p": -0.25, "r": 0, "s": 0, "z": -0.2}[name[-1]]
        # direction = self.direction
        direction = wrapper.direction or _enums.DOWN
        direction = _string.to_tridirectional_lilypond_symbol(direction)
        strings = []
        strings.append(f"{direction} #(make-dynamic-script")
        strings.append("    (markup")
        strings.append("        #:whiteout")
        strings.append("        #:line (")
        strings.append('            #:general-align Y -2 #:normal-text #:larger "“"')
        strings.append(f"            #:hspace {before}")
        strings.append(f'            #:dynamic "{name}"')
        strings.append(f"            #:hspace {after}")
        strings.append('            #:general-align Y -2 #:normal-text #:larger "”"')
        strings.append("            )")
        strings.append("        )")
        strings.append("    )")
        string = "\n".join(strings)
        return string

    @staticmethod
    # def _format_textual(direction, string):
    def _format_textual(string, *, wrapper=None):
        if wrapper.direction is None:
            direction = _enums.DOWN
        direction = _string.to_tridirectional_lilypond_symbol(direction)
        assert isinstance(string, str), repr(string)
        string = f'(markup #:whiteout #:normal-text #:italic "{string}")'
        string = f"{direction} #(make-dynamic-script {string})"
        return string

    def _get_lilypond_format(self, *, wrapper=None):
        if self.command:
            string = self.command
        elif self.effort:
            string = self._format_effort_dynamic(wrapper=wrapper)
        elif self.name_is_textual:
            string = self._format_textual(self.name, wrapper=wrapper)
        else:
            string = rf"\{self.name}"
            if wrapper.direction is not None:
                direction_ = wrapper.direction
                direction = _string.to_tridirectional_lilypond_symbol(direction_)
                string = f"{direction} {string}"
        return string

    def _get_contributions(self, *, component=None, wrapper=None):
        contributions = _contributions.ContributionsBySite()
        command = self._get_lilypond_format(wrapper=wrapper)
        if self.leak:
            contributions.after.leak.append(_EMPTY_CHORD)
            if not self.hide:
                contributions.after.leaks.append(command)
        else:
            if not self.hide:
                contributions.after.articulations.append(command)
        return contributions

    # TODO: make markup function and shorten docstring
    @property
    def effort(self) -> bool:
        r"""
        Is true when double quotes enclose dynamic.

        ..  container:: example

            >>> voice = abjad.Voice("c'4 r d' r e' r f' r")
            >>> abjad.attach(abjad.Dynamic('"pp"'), voice[0])
            >>> abjad.attach(abjad.Dynamic('"mp"'), voice[2])
            >>> abjad.attach(abjad.Dynamic('"mf"'), voice[4])
            >>> abjad.attach(abjad.Dynamic('"ff"'), voice[6])
            >>> abjad.override(voice).DynamicLineSpanner.staff_padding = 4
            >>> abjad.show(voice) # doctest: +SKIP

            ..  docs::

                >>> string = abjad.lilypond(voice)
                >>> print(string)
                \new Voice
                \with
                {
                    \override DynamicLineSpanner.staff-padding = 4
                }
                {
                    c'4
                    _ #(make-dynamic-script
                        (markup
                            #:whiteout
                            #:line (
                                #:general-align Y -2 #:normal-text #:larger "“"
                                #:hspace -0.1
                                #:dynamic "pp"
                                #:hspace -0.25
                                #:general-align Y -2 #:normal-text #:larger "”"
                                )
                            )
                        )
                    r4
                    d'4
                    _ #(make-dynamic-script
                        (markup
                            #:whiteout
                            #:line (
                                #:general-align Y -2 #:normal-text #:larger "“"
                                #:hspace -0.1
                                #:dynamic "mp"
                                #:hspace -0.25
                                #:general-align Y -2 #:normal-text #:larger "”"
                                )
                            )
                        )
                    r4
                    e'4
                    _ #(make-dynamic-script
                        (markup
                            #:whiteout
                            #:line (
                                #:general-align Y -2 #:normal-text #:larger "“"
                                #:hspace -0.1
                                #:dynamic "mf"
                                #:hspace -0.2
                                #:general-align Y -2 #:normal-text #:larger "”"
                                )
                            )
                        )
                    r4
                    f'4
                    _ #(make-dynamic-script
                        (markup
                            #:whiteout
                            #:line (
                                #:general-align Y -2 #:normal-text #:larger "“"
                                #:hspace -0.4
                                #:dynamic "ff"
                                #:hspace -0.2
                                #:general-align Y -2 #:normal-text #:larger "”"
                                )
                            )
                        )
                    r4
                }

        ..  container:: example

            >>> voice = abjad.Voice("c'4 r d' r e' r f' r")
            >>> abjad.attach(abjad.Dynamic('"sf"'), voice[0])
            >>> abjad.attach(abjad.Dynamic('"sfz"'), voice[2])
            >>> abjad.attach(abjad.Dynamic('"rf"'), voice[4])
            >>> abjad.attach(abjad.Dynamic('"rfz"'), voice[6])
            >>> abjad.override(voice).DynamicLineSpanner.staff_padding = 4
            >>> abjad.show(voice) # doctest: +SKIP

            ..  docs::

                >>> string = abjad.lilypond(voice)
                >>> print(string)
                \new Voice
                \with
                {
                    \override DynamicLineSpanner.staff-padding = 4
                }
                {
                    c'4
                    _ #(make-dynamic-script
                        (markup
                            #:whiteout
                            #:line (
                                #:general-align Y -2 #:normal-text #:larger "“"
                                #:hspace -0.3
                                #:dynamic "sf"
                                #:hspace -0.2
                                #:general-align Y -2 #:normal-text #:larger "”"
                                )
                            )
                        )
                    r4
                    d'4
                    _ #(make-dynamic-script
                        (markup
                            #:whiteout
                            #:line (
                                #:general-align Y -2 #:normal-text #:larger "“"
                                #:hspace -0.3
                                #:dynamic "sfz"
                                #:hspace -0.2
                                #:general-align Y -2 #:normal-text #:larger "”"
                                )
                            )
                        )
                    r4
                    e'4
                    _ #(make-dynamic-script
                        (markup
                            #:whiteout
                            #:line (
                                #:general-align Y -2 #:normal-text #:larger "“"
                                #:hspace -0.1
                                #:dynamic "rf"
                                #:hspace -0.2
                                #:general-align Y -2 #:normal-text #:larger "”"
                                )
                            )
                        )
                    r4
                    f'4
                    _ #(make-dynamic-script
                        (markup
                            #:whiteout
                            #:line (
                                #:general-align Y -2 #:normal-text #:larger "“"
                                #:hspace -0.1
                                #:dynamic "rfz"
                                #:hspace -0.2
                                #:general-align Y -2 #:normal-text #:larger "”"
                                )
                            )
                        )
                    r4
                }

        """
        assert isinstance(self.name, str), repr(self.name)
        return bool(self.name) and self.name[0] == '"'

    @property
    def sforzando(self) -> bool:
        """
        Is true when dynamic name begins in s- and ends in -z.

        ..  container:: example

            >>> abjad.Dynamic("f").sforzando
            False

            >>> abjad.Dynamic("sfz").sforzando
            True

            >>> abjad.Dynamic("sffz").sforzando
            True

            >>> abjad.Dynamic("sfp").sforzando
            False

            >>> abjad.Dynamic("sf").sforzando
            False

            >>> abjad.Dynamic("rfz").sforzando
            False

        """
        assert isinstance(self.name, str), repr(self.name)
        if self.name and self.name.startswith("s") and self.name.endswith("z"):
            return True
        return False

    # TODO: make module-level function
    @staticmethod
    def composite_dynamic_name_to_steady_state_dynamic_name(name) -> str:
        """
        Changes composite ``name`` to steady state dynamic name.

        ..  container:: example

            >>> abjad.Dynamic.composite_dynamic_name_to_steady_state_dynamic_name("sfp")
            'p'

            >>> abjad.Dynamic.composite_dynamic_name_to_steady_state_dynamic_name("rfz")
            'f'

        """
        return Dynamic._composite_dynamic_name_to_steady_state_dynamic_name[name]

    # TODO: make module-level function
    @staticmethod
    def dynamic_name_to_dynamic_ordinal(name):
        """
        Changes ``name`` to dynamic ordinal.

        ..  container:: example

            >>> abjad.Dynamic.dynamic_name_to_dynamic_ordinal("fff")
            4

            >>> abjad.Dynamic.dynamic_name_to_dynamic_ordinal("niente")
            NegativeInfinity()

        """
        try:
            return Dynamic._dynamic_name_to_dynamic_ordinal[name]
        except KeyError:
            name = Dynamic.composite_dynamic_name_to_steady_state_dynamic_name(name)
            return Dynamic._dynamic_name_to_dynamic_ordinal[name]

    # TODO: make module-level function
    @staticmethod
    def dynamic_ordinal_to_dynamic_name(dynamic_ordinal) -> str:
        """
        Changes ``dynamic_ordinal`` to dynamic name.

        ..  container:: example

            >>> abjad.Dynamic.dynamic_ordinal_to_dynamic_name(-5)
            'pppp'

            >>> negative_infinity = abjad.math.NegativeInfinity()
            >>> abjad.Dynamic.dynamic_ordinal_to_dynamic_name(negative_infinity)
            'niente'

        """
        if dynamic_ordinal == _math.NegativeInfinity():
            return "niente"
        else:
            return Dynamic._dynamic_ordinal_to_dynamic_name[dynamic_ordinal]

    # TODO: make module-level function
    @staticmethod
    def is_dynamic_name(argument) -> bool:
        """
        Is true when ``argument`` is dynamic name.

        ..  container:: example

            >>> abjad.Dynamic.is_dynamic_name("f")
            True

            >>> abjad.Dynamic.is_dynamic_name("sfz")
            True

            >>> abjad.Dynamic.is_dynamic_name("niente")
            True

        """
        return argument in Dynamic._dynamic_names



r"""
Instruments.

..  container:: example

    Default ranges of each instrument:

    >>> names = [_ for _ in dir(abjad.instruments) if _[0].isupper()]
    >>> for name in names:
    ...     class_ = abjad.__dict__[name]
    ...     if issubclass(class_, abjad.Instrument):
    ...         instrument = class_()
    ...         instrument_name = instrument.__class__.__name__
    ...         range_string = instrument.pitch_range.range_string
    ...         print(f"{instrument_name}: {range_string}")
    Accordion: [E1, C8]
    AltoFlute: [G3, G6]
    AltoSaxophone: [Db3, A5]
    AltoTrombone: [A2, Bb5]
    AltoVoice: [F3, G5]
    BaritoneSaxophone: [C2, Ab4]
    BaritoneVoice: [A2, A4]
    BassClarinet: [Bb1, G5]
    BassFlute: [C3, C6]
    BassSaxophone: [Ab2, E4]
    BassTrombone: [C2, F4]
    BassVoice: [E2, F4]
    Bassoon: [Bb1, Eb5]
    Cello: [C2, G5]
    ClarinetInA: [Db3, A6]
    ClarinetInBFlat: [D3, Bb6]
    ClarinetInEFlat: [F3, C7]
    Contrabass: [C1, G4]
    ContrabassClarinet: [Bb0, G4]
    ContrabassFlute: [G2, G5]
    ContrabassSaxophone: [C1, Ab3]
    Contrabassoon: [Bb0, Bb4]
    EnglishHorn: [E3, C6]
    Flute: [C4, D7]
    FrenchHorn: [B1, F5]
    Glockenspiel: [G5, C8]
    Guitar: [E2, E5]
    Harp: [B0, G#7]
    Harpsichord: [C2, C7]
    Instrument: [-inf, +inf]
    Marimba: [F2, C7]
    MezzoSopranoVoice: [A3, C6]
    Oboe: [Bb3, A6]
    Percussion: [-inf, +inf]
    Piano: [A0, C8]
    Piccolo: [D5, C8]
    SopraninoSaxophone: [Db4, F#6]
    SopranoSaxophone: [Ab3, E6]
    SopranoVoice: [C4, E6]
    TenorSaxophone: [Ab2, E5]
    TenorTrombone: [E2, Eb5]
    TenorVoice: [C3, D5]
    Trumpet: [F#3, D6]
    Tuba: [D1, F4]
    Vibraphone: [F3, F6]
    Viola: [C3, D6]
    Violin: [G3, G7]
    Xylophone: [C4, C7]

..  container:: example

    Two instruments active on a single staff:

    >>> voice_1 = abjad.Voice("e'8 g'8 f'8 a'8")
    >>> flute = abjad.Flute()
    >>> abjad.attach(flute, voice_1[0], context="Voice")
    >>> abjad.attach(abjad.LilyPondLiteral(r"\voiceOne"), voice_1)
    >>> voice_2 = abjad.Voice("c'2")
    >>> abjad.attach(abjad.LilyPondLiteral(r"\voiceTwo"), voice_2)
    >>> viola = abjad.Viola()
    >>> abjad.attach(viola, voice_2[0], context="Voice")
    >>> staff = abjad.Staff([voice_1, voice_2], simultaneous=True)
    >>> abjad.show(staff) # doctest: +SKIP

    ..  docs::

        >>> string = abjad.lilypond(staff)
        >>> print(string)
        \new Staff
        <<
            \new Voice
            {
                \voiceOne
                e'8
                g'8
                f'8
                a'8
            }
            \new Voice
            {
                \voiceTwo
                c'2
            }
        >>

    >>> for leaf in abjad.select.leaves(voice_1):
    ...     leaf, abjad.get.effective(leaf, abjad.Instrument)
    ...
    (Note("e'8"), Flute(clefs=('treble',), context='Staff', middle_c_sounding_pitch=NamedPitch("c'"), pitch_range=PitchRange(range_string='[C4, D7]')))
    (Note("g'8"), Flute(clefs=('treble',), context='Staff', middle_c_sounding_pitch=NamedPitch("c'"), pitch_range=PitchRange(range_string='[C4, D7]')))
    (Note("f'8"), Flute(clefs=('treble',), context='Staff', middle_c_sounding_pitch=NamedPitch("c'"), pitch_range=PitchRange(range_string='[C4, D7]')))
    (Note("a'8"), Flute(clefs=('treble',), context='Staff', middle_c_sounding_pitch=NamedPitch("c'"), pitch_range=PitchRange(range_string='[C4, D7]')))

    >>> for leaf in abjad.select.leaves(voice_2):
    ...     leaf, abjad.get.effective(leaf, abjad.Instrument)
    ...
    (Note("c'2"), Viola(clefs=('alto', 'treble'), context='Staff', middle_c_sounding_pitch=NamedPitch("c'"), pitch_range=PitchRange(range_string='[C3, D6]'), tuning=Tuning(pitches=(NamedPitch('c'), NamedPitch('g'), NamedPitch("d'"), NamedPitch("a'")))))

"""
import dataclasses
import typing

from . import contributions as _contributions
from . import enumerate as _enumerate
from . import pcollections as _pcollections
from . import pitch as _pitch


@dataclasses.dataclass(frozen=True, order=True, slots=True, unsafe_hash=True)
class Instrument:
    """
    Instrument.
    """

    clefs: tuple[str, ...] = ()
    context: str = "Staff"
    middle_c_sounding_pitch: _pitch.NamedPitch = _pitch.NamedPitch("C4")
    pitch_range: _pcollections.PitchRange = _pcollections.PitchRange("[-inf, +inf]")

    _site: typing.ClassVar[str] = "opening"
    _latent: typing.ClassVar[bool] = True
    _persistent: typing.ClassVar[bool] = True
    _redraw: typing.ClassVar[bool] = True
    check_effective_context: typing.ClassVar[bool] = True

    def __post_init__(self):
        assert isinstance(self.context, str), repr(self.context)
        assert isinstance(self.clefs, tuple), repr(self.clefs)
        assert all(isinstance(_, str) for _ in self.clefs)
        assert isinstance(self.pitch_range, _pcollections.PitchRange), repr(
            self.pitch_range
        )
        assert isinstance(self.middle_c_sounding_pitch, _pitch.NamedPitch), repr(
            self.middle_c_sounding_pitch
        )

    @property
    def _lilypond_type(self):
        if isinstance(self.context, type):
            return self.context.__name__
        elif isinstance(self.context, str):
            return self.context
        else:
            return type(self.context).__name__

    def _attachment_test_all(self, leaf):
        assert hasattr(leaf, "written_duration")
        if leaf._has_indicator(Instrument):
            string = f"Already has instrument: {leaf}."
            return string
        return True

    def _get_contributions(self, *, component=None, wrapper=None):
        contributions = _contributions.ContributionsBySite()
        strings = self._get_lilypond_format()
        assert isinstance(strings, list), repr(strings)
        contributions.opening.commands.extend(strings)
        return contributions

    def _get_lilypond_format(self, context=None):
        return []

    @property
    def latent(self) -> bool:
        """
        Is true for all instruments.

        Class constant.
        """
        return self._latent

    @property
    def persistent(self) -> bool:
        """
        Is true.

        Class constant.
        """
        return self._persistent

    @property
    def redraw(self) -> bool:
        """
        Is true for all instruments.

        Class constant.
        """
        return self._redraw


@dataclasses.dataclass(frozen=True, order=True, slots=True, unsafe_hash=True)
class StringNumber:
    """
    String number.

    ..  container:: example

        String I:

        >>> abjad.StringNumber((1,))
        StringNumber(numbers=(1,))

        Strings II and III:

        >>> abjad.StringNumber((2, 3))
        StringNumber(numbers=(2, 3))

    """

    numbers: tuple[int, ...] = (1,)

    def __post_init__(self):
        assert isinstance(self.numbers, tuple), repr(self.numbers)
        assert all(0 < _ < 7 for _ in self.numbers)

    @property
    def roman_numerals(self) -> tuple[str, ...]:
        """
        Gets roman numerals of string number indicator.

        ..  container:: example

            String I:

            >>> indicator = abjad.StringNumber((1,))
            >>> indicator.roman_numerals
            ('i',)

            Strings II and III:

            >>> indicator = abjad.StringNumber((2, 3))
            >>> indicator.roman_numerals
            ('ii', 'iii')

        """
        numerals = ("i", "ii", "iii", "iv", "v", "vi")
        result = []
        for number in self.numbers:
            numeral = numerals[number - 1]
            result.append(numeral)
        return tuple(result)


@dataclasses.dataclass(frozen=True, order=True, slots=True, unsafe_hash=True)
class Tuning:
    """
    Tuning.

    ..  container:: example

        Violin tuning:

        >>> pitches = [abjad.NamedPitch(_) for _ in "G3 D4 A4 E5".split()]
        >>> abjad.Tuning(tuple(pitches))
        Tuning(pitches=(NamedPitch('g'), NamedPitch("d'"), NamedPitch("a'"), NamedPitch("e''")))

    """

    pitches: tuple[_pitch.NamedPitch, ...] = ()

    def __post_init__(self):
        assert isinstance(self.pitches, tuple), repr(self.pitches)
        assert all(isinstance(_, _pitch.NamedPitch) for _ in self.pitches)

    @property
    def pitch_ranges(self) -> list[_pcollections.PitchRange]:
        """
        Gets two-octave pitch-ranges for each pitch in this tuning.

        ..  container:: example

            >>> pitches = [abjad.NamedPitch(_) for _ in "G3 D4 A4 E5".split()]
            >>> tuning = abjad.Tuning(tuple(pitches))
            >>> for range_ in tuning.pitch_ranges:
            ...     range_
            PitchRange(range_string='[G3, G5]')
            PitchRange(range_string='[D4, D6]')
            PitchRange(range_string='[A4, A6]')
            PitchRange(range_string='[E5, E7]')

        """
        result = []
        for pitch in self.pitches or []:
            pitch_range: _pcollections.PitchRange = _pcollections.PitchRange(
                f"[{pitch.name}, {(pitch + 24).name}]"
            )
            result.append(pitch_range)
        return result

    def get_pitch_ranges_by_string_number(
        self, string_number: StringNumber
    ) -> tuple[_pcollections.PitchRange, ...]:
        """
        Gets tuning pitch ranges by string number.

        ..  container:: example

            Violin tuning:

            >>> pitches = [abjad.NamedPitch(_) for _ in "G3 D4 A4 E5".split()]
            >>> tuning = abjad.Tuning(tuple(pitches))
            >>> string_number = abjad.StringNumber((2, 3))
            >>> tuning.get_pitch_ranges_by_string_number(string_number)
            (PitchRange(range_string='[A4, A6]'), PitchRange(range_string='[D4, D6]'))

        """
        if not isinstance(string_number, StringNumber):
            string_number = StringNumber(string_number)
        assert isinstance(string_number, StringNumber)
        pitch_ranges = self.pitch_ranges
        result = []
        for number in string_number.numbers:
            index = -number
            pitch_range: _pcollections.PitchRange = pitch_ranges[index]
            result.append(pitch_range)
        return tuple(result)

    def get_pitches_by_string_number(
        self, string_number: StringNumber
    ) -> tuple[_pitch.NamedPitch, ...]:
        """
        Gets tuning pitches by string number.

        ..  container:: example

            Violin tuning:

            >>> pitches = [abjad.NamedPitch(_) for _ in "G3 D4 A4 E5".split()]
            >>> tuning = abjad.Tuning(tuple(pitches))
            >>> string_number = abjad.StringNumber((2, 3))
            >>> tuning.get_pitches_by_string_number(string_number)
            (NamedPitch("a'"), NamedPitch("d'"))

        """
        if not isinstance(string_number, StringNumber):
            string_number = StringNumber(string_number)
        assert isinstance(string_number, StringNumber)
        assert self.pitches is not None
        result = []
        for number in string_number.numbers:
            index = -number
            pitch = self.pitches[index]
            result.append(pitch)
        return tuple(result)

    def voice_pitch_classes(self, pitch_classes, allow_open_strings: bool = True):
        r"""
        Voices ``pitch_classes``.

        ..  container:: example

            >>> pitches = [abjad.NamedPitch(_) for _ in "G3 D4 A4 E5".split()]
            >>> tuning = abjad.Tuning(tuple(pitches))
            >>> voicings = tuning.voice_pitch_classes(('a',))
            >>> for voicing in voicings:
            ...     voicing
            ...
            (NamedPitch('a'), None, None, None)
            (None, None, None, NamedPitch("a''"))
            (None, None, None, NamedPitch("a'''"))
            (None, None, NamedPitch("a'"), None)
            (None, None, NamedPitch("a''"), None)
            (None, None, NamedPitch("a'''"), None)
            (None, NamedPitch("a'"), None, None)
            (None, NamedPitch("a''"), None, None)
            (NamedPitch("a'"), None, None, None)

            >>> voicings = tuning.voice_pitch_classes(
            ...     ("a", "d"),
            ...     allow_open_strings=False,
            ... )
            >>> for voicing in voicings:
            ...     voicing
            ...
            (NamedPitch('a'), None, None, NamedPitch("d'''"))
            (NamedPitch('a'), None, None, NamedPitch("d''''"))
            (NamedPitch('a'), None, NamedPitch("d''"), None)
            (NamedPitch('a'), None, NamedPitch("d'''"), None)
            (NamedPitch('a'), NamedPitch("d''"), None, None)
            (NamedPitch('a'), NamedPitch("d'''"), None, None)
            (None, None, NamedPitch("d''"), NamedPitch("a''"))
            (None, None, NamedPitch("d''"), NamedPitch("a'''"))
            (None, None, NamedPitch("a''"), NamedPitch("d'''"))
            (None, None, NamedPitch("a''"), NamedPitch("d''''"))
            (None, None, NamedPitch("d'''"), NamedPitch("a''"))
            (None, None, NamedPitch("d'''"), NamedPitch("a'''"))
            (None, None, NamedPitch("a'''"), NamedPitch("d'''"))
            (None, None, NamedPitch("a'''"), NamedPitch("d''''"))
            (None, NamedPitch("a'"), None, NamedPitch("d'''"))
            (None, NamedPitch("a'"), None, NamedPitch("d''''"))
            (None, NamedPitch("a'"), NamedPitch("d''"), None)
            (None, NamedPitch("a'"), NamedPitch("d'''"), None)
            (None, NamedPitch("d''"), None, NamedPitch("a''"))
            (None, NamedPitch("d''"), None, NamedPitch("a'''"))
            (None, NamedPitch("d''"), NamedPitch("a''"), None)
            (None, NamedPitch("d''"), NamedPitch("a'''"), None)
            (None, NamedPitch("a''"), None, NamedPitch("d'''"))
            (None, NamedPitch("a''"), None, NamedPitch("d''''"))
            (None, NamedPitch("a''"), NamedPitch("d''"), None)
            (None, NamedPitch("a''"), NamedPitch("d'''"), None)
            (None, NamedPitch("d'''"), None, NamedPitch("a''"))
            (None, NamedPitch("d'''"), None, NamedPitch("a'''"))
            (None, NamedPitch("d'''"), NamedPitch("a''"), None)
            (None, NamedPitch("d'''"), NamedPitch("a'''"), None)
            (NamedPitch("d'"), None, None, NamedPitch("a''"))
            (NamedPitch("d'"), None, None, NamedPitch("a'''"))
            (NamedPitch("d'"), None, NamedPitch("a''"), None)
            (NamedPitch("d'"), None, NamedPitch("a'''"), None)
            (NamedPitch("d'"), NamedPitch("a'"), None, None)
            (NamedPitch("d'"), NamedPitch("a''"), None, None)
            (NamedPitch("a'"), None, None, NamedPitch("d'''"))
            (NamedPitch("a'"), None, None, NamedPitch("d''''"))
            (NamedPitch("a'"), None, NamedPitch("d''"), None)
            (NamedPitch("a'"), None, NamedPitch("d'''"), None)
            (NamedPitch("a'"), NamedPitch("d''"), None, None)
            (NamedPitch("a'"), NamedPitch("d'''"), None, None)
            (NamedPitch("d''"), None, None, NamedPitch("a''"))
            (NamedPitch("d''"), None, None, NamedPitch("a'''"))
            (NamedPitch("d''"), None, NamedPitch("a''"), None)
            (NamedPitch("d''"), None, NamedPitch("a'''"), None)
            (NamedPitch("d''"), NamedPitch("a'"), None, None)
            (NamedPitch("d''"), NamedPitch("a''"), None, None)

        """
        assert self.pitches is not None
        pitch_classes = [_pitch.NamedPitchClass(_) for _ in pitch_classes]
        pitch_classes.extend([None] * (len(self.pitches) - len(pitch_classes)))
        permutations = _enumerate.yield_permutations(pitch_classes)
        permutations = set([tuple(_) for _ in permutations])
        pitch_ranges = self.pitch_ranges
        result: list[tuple[_pitch.NamedPitch | None, ...]] = []
        for permutation in permutations:
            sequences: list = []
            for pitch_range, pitch_class in zip(pitch_ranges, permutation):
                pitches: list[_pitch.NamedPitch | None]
                if pitch_class is None:
                    sequences.append([None])
                    continue
                pitches = list(pitch_range.voice_pitch_class(pitch_class))
                if not allow_open_strings:
                    pitches = [
                        pitch for pitch in pitches if pitch != pitch_range.start_pitch
                    ]
                if not pitches:
                    pitches = [None]
                sequences.append(pitches)
            subresult = _enumerate.outer_product(sequences)
            subresult = [tuple(x) for x in subresult]
            result.extend(subresult)
        result.sort()
        return tuple(result)


@dataclasses.dataclass(frozen=True, order=True, slots=True, unsafe_hash=True)
class Accordion(Instrument):

    clefs: tuple[str, ...] = ("treble", "bass")
    context: str = "StaffGroup"
    middle_c_sounding_pitch: _pitch.NamedPitch = _pitch.NamedPitch("C4")
    pitch_range: _pcollections.PitchRange = _pcollections.PitchRange("[E1, C8]")


@dataclasses.dataclass(frozen=True, order=True, slots=True, unsafe_hash=True)
class AltoFlute(Instrument):

    clefs: tuple[str, ...] = ("treble",)
    context: str = "Staff"
    middle_c_sounding_pitch: _pitch.NamedPitch = _pitch.NamedPitch("G3")
    pitch_range: _pcollections.PitchRange = _pcollections.PitchRange("[G3, G6]")


@dataclasses.dataclass(frozen=True, order=True, slots=True, unsafe_hash=True)
class AltoSaxophone(Instrument):

    clefs: tuple[str, ...] = ("treble",)
    context: str = "Staff"
    middle_c_sounding_pitch: _pitch.NamedPitch = _pitch.NamedPitch("Eb3")
    pitch_range: _pcollections.PitchRange = _pcollections.PitchRange("[Db3, A5]")


@dataclasses.dataclass(frozen=True, order=True, slots=True, unsafe_hash=True)
class AltoTrombone(Instrument):

    clefs: tuple[str, ...] = ("bass", "tenor")
    context: str = "Staff"
    middle_c_sounding_pitch: _pitch.NamedPitch = _pitch.NamedPitch("C4")
    pitch_range: _pcollections.PitchRange = _pcollections.PitchRange("[A2, Bb5]")


@dataclasses.dataclass(frozen=True, order=True, slots=True, unsafe_hash=True)
class AltoVoice(Instrument):

    clefs: tuple[str, ...] = ("treble",)
    context: str = "Staff"
    middle_c_sounding_pitch: _pitch.NamedPitch = _pitch.NamedPitch("C4")
    pitch_range: _pcollections.PitchRange = _pcollections.PitchRange("[F3, G5]")


@dataclasses.dataclass(frozen=True, order=True, slots=True, unsafe_hash=True)
class BaritoneSaxophone(Instrument):

    clefs: tuple[str, ...] = ("treble",)
    context: str = "Staff"
    middle_c_sounding_pitch: _pitch.NamedPitch = _pitch.NamedPitch("Eb2")
    pitch_range: _pcollections.PitchRange = _pcollections.PitchRange("[C2, Ab4]")


@dataclasses.dataclass(frozen=True, order=True, slots=True, unsafe_hash=True)
class BaritoneVoice(Instrument):

    clefs: tuple[str, ...] = ("bass",)
    context: str = "Staff"
    middle_c_sounding_pitch: _pitch.NamedPitch = _pitch.NamedPitch("C4")
    pitch_range: _pcollections.PitchRange = _pcollections.PitchRange("[A2, A4]")


@dataclasses.dataclass(frozen=True, order=True, slots=True, unsafe_hash=True)
class BassClarinet(Instrument):

    clefs: tuple[str, ...] = ("treble", "bass")
    context: str = "Staff"
    middle_c_sounding_pitch: _pitch.NamedPitch = _pitch.NamedPitch("Bb2")
    pitch_range: _pcollections.PitchRange = _pcollections.PitchRange("[Bb1, G5]")


@dataclasses.dataclass(frozen=True, order=True, slots=True, unsafe_hash=True)
class BassFlute(Instrument):

    clefs: tuple[str, ...] = ("treble",)
    context: str = "Staff"
    middle_c_sounding_pitch: _pitch.NamedPitch = _pitch.NamedPitch("C3")
    pitch_range: _pcollections.PitchRange = _pcollections.PitchRange("[C3, C6]")


@dataclasses.dataclass(frozen=True, order=True, slots=True, unsafe_hash=True)
class BassSaxophone(Instrument):

    clefs: tuple[str, ...] = ("treble",)
    context: str = "Staff"
    middle_c_sounding_pitch: _pitch.NamedPitch = _pitch.NamedPitch("Bb1")
    pitch_range: _pcollections.PitchRange = _pcollections.PitchRange("[Ab2, E4]")


@dataclasses.dataclass(frozen=True, order=True, slots=True, unsafe_hash=True)
class BassTrombone(Instrument):

    clefs: tuple[str, ...] = ("bass",)
    context: str = "Staff"
    middle_c_sounding_pitch: _pitch.NamedPitch = _pitch.NamedPitch("C4")
    pitch_range: _pcollections.PitchRange = _pcollections.PitchRange("[C2, F4]")


@dataclasses.dataclass(frozen=True, order=True, slots=True, unsafe_hash=True)
class BassVoice(Instrument):

    clefs: tuple[str, ...] = ("bass",)
    context: str = "Staff"
    middle_c_sounding_pitch: _pitch.NamedPitch = _pitch.NamedPitch("C4")
    pitch_range: _pcollections.PitchRange = _pcollections.PitchRange("[E2, F4]")


@dataclasses.dataclass(frozen=True, order=True, slots=True, unsafe_hash=True)
class Bassoon(Instrument):

    clefs: tuple[str, ...] = ("bass", "tenor")
    context: str = "Staff"
    middle_c_sounding_pitch: _pitch.NamedPitch = _pitch.NamedPitch("C4")
    pitch_range: _pcollections.PitchRange = _pcollections.PitchRange("[Bb1, Eb5]")


@dataclasses.dataclass(frozen=True, order=True, slots=True, unsafe_hash=True)
class Cello(Instrument):

    clefs: tuple[str, ...] = ("bass", "tenor", "treble")
    context: str = "Staff"
    middle_c_sounding_pitch: _pitch.NamedPitch = _pitch.NamedPitch("C4")
    pitch_range: _pcollections.PitchRange = _pcollections.PitchRange("[C2, G5]")
    tuning: Tuning = Tuning(
        tuple([_pitch.NamedPitch(_) for _ in "C2 G2 D3 A3".split()])
    )


@dataclasses.dataclass(frozen=True, order=True, slots=True, unsafe_hash=True)
class ClarinetInA(Instrument):

    clefs: tuple[str, ...] = ("treble",)
    context: str = "Staff"
    middle_c_sounding_pitch: _pitch.NamedPitch = _pitch.NamedPitch("A3")
    pitch_range: _pcollections.PitchRange = _pcollections.PitchRange("[Db3, A6]")


@dataclasses.dataclass(frozen=True, order=True, slots=True, unsafe_hash=True)
class ClarinetInBFlat(Instrument):

    clefs: tuple[str, ...] = ("treble",)
    context: str = "Staff"
    middle_c_sounding_pitch: _pitch.NamedPitch = _pitch.NamedPitch("Bb3")
    pitch_range: _pcollections.PitchRange = _pcollections.PitchRange("[D3, Bb6]")


@dataclasses.dataclass(frozen=True, order=True, slots=True, unsafe_hash=True)
class ClarinetInEFlat(Instrument):

    clefs: tuple[str, ...] = ("treble",)
    context: str = "Staff"
    middle_c_sounding_pitch: _pitch.NamedPitch = _pitch.NamedPitch("Eb4")
    pitch_range: _pcollections.PitchRange = _pcollections.PitchRange("[F3, C7]")


@dataclasses.dataclass(frozen=True, order=True, slots=True, unsafe_hash=True)
class Contrabass(Instrument):

    clefs: tuple[str, ...] = ("bass", "treble")
    context: str = "Staff"
    middle_c_sounding_pitch: _pitch.NamedPitch = _pitch.NamedPitch("C3")
    pitch_range: _pcollections.PitchRange = _pcollections.PitchRange("[C1, G4]")
    tuning: Tuning = Tuning(
        tuple([_pitch.NamedPitch(_) for _ in "C1 A1 D2 G2".split()])
    )


@dataclasses.dataclass(frozen=True, order=True, slots=True, unsafe_hash=True)
class ContrabassClarinet(Instrument):

    clefs: tuple[str, ...] = ("treble", "bass")
    context: str = "Staff"
    middle_c_sounding_pitch: _pitch.NamedPitch = _pitch.NamedPitch("Bb1")
    pitch_range: _pcollections.PitchRange = _pcollections.PitchRange("[Bb0, G4]")


@dataclasses.dataclass(frozen=True, order=True, slots=True, unsafe_hash=True)
class ContrabassFlute(Instrument):

    clefs: tuple[str, ...] = ("treble",)
    context: str = "Staff"
    middle_c_sounding_pitch: _pitch.NamedPitch = _pitch.NamedPitch("G2")
    pitch_range: _pcollections.PitchRange = _pcollections.PitchRange("[G2, G5]")


@dataclasses.dataclass(frozen=True, order=True, slots=True, unsafe_hash=True)
class ContrabassSaxophone(Instrument):

    clefs: tuple[str, ...] = ("treble",)
    context: str = "Staff"
    middle_c_sounding_pitch: _pitch.NamedPitch = _pitch.NamedPitch("Eb1")
    pitch_range: _pcollections.PitchRange = _pcollections.PitchRange("[C1, Ab3]")


@dataclasses.dataclass(frozen=True, order=True, slots=True, unsafe_hash=True)
class Contrabassoon(Instrument):

    clefs: tuple[str, ...] = ("bass",)
    context: str = "Staff"
    middle_c_sounding_pitch: _pitch.NamedPitch = _pitch.NamedPitch("C3")
    pitch_range: _pcollections.PitchRange = _pcollections.PitchRange("[Bb0, Bb4]")


@dataclasses.dataclass(frozen=True, order=True, slots=True, unsafe_hash=True)
class EnglishHorn(Instrument):

    clefs: tuple[str, ...] = ("treble",)
    context: str = "Staff"
    middle_c_sounding_pitch: _pitch.NamedPitch = _pitch.NamedPitch("F3")
    pitch_range: _pcollections.PitchRange = _pcollections.PitchRange("[E3, C6]")


@dataclasses.dataclass(frozen=True, order=True, slots=True, unsafe_hash=True)
class Flute(Instrument):

    clefs: tuple[str, ...] = ("treble",)
    context: str = "Staff"
    middle_c_sounding_pitch: _pitch.NamedPitch = _pitch.NamedPitch("C4")
    pitch_range: _pcollections.PitchRange = _pcollections.PitchRange("[C4, D7]")


@dataclasses.dataclass(frozen=True, order=True, slots=True, unsafe_hash=True)
class FrenchHorn(Instrument):

    clefs: tuple[str, ...] = ("bass", "treble")
    context: str = "Staff"
    middle_c_sounding_pitch: _pitch.NamedPitch = _pitch.NamedPitch("F3")
    pitch_range: _pcollections.PitchRange = _pcollections.PitchRange("[B1, F5]")


@dataclasses.dataclass(frozen=True, order=True, slots=True, unsafe_hash=True)
class Glockenspiel(Instrument):

    clefs: tuple[str, ...] = ("treble",)
    context: str = "Staff"
    middle_c_sounding_pitch: _pitch.NamedPitch = _pitch.NamedPitch("C6")
    pitch_range: _pcollections.PitchRange = _pcollections.PitchRange("[G5, C8]")


@dataclasses.dataclass(frozen=True, order=True, slots=True, unsafe_hash=True)
class Guitar(Instrument):

    clefs: tuple[str, ...] = ("treble",)
    context: str = "Staff"
    middle_c_sounding_pitch: _pitch.NamedPitch = _pitch.NamedPitch("C3")
    pitch_range: _pcollections.PitchRange = _pcollections.PitchRange("[E2, E5]")
    tuning: Tuning = Tuning(
        tuple([_pitch.NamedPitch(_) for _ in "E2 A2 D3 G3 B3 E4".split()])
    )


@dataclasses.dataclass(frozen=True, order=True, slots=True, unsafe_hash=True)
class Harp(Instrument):

    clefs: tuple[str, ...] = ("treble", "bass")
    context: str = "StaffGroup"
    middle_c_sounding_pitch: _pitch.NamedPitch = _pitch.NamedPitch("C4")
    pitch_range: _pcollections.PitchRange = _pcollections.PitchRange("[B0, G#7]")


@dataclasses.dataclass(frozen=True, order=True, slots=True, unsafe_hash=True)
class Harpsichord(Instrument):

    clefs: tuple[str, ...] = ("treble", "bass")
    context: str = "StaffGroup"
    middle_c_sounding_pitch: _pitch.NamedPitch = _pitch.NamedPitch("C4")
    pitch_range: _pcollections.PitchRange = _pcollections.PitchRange("[C2, C7]")


@dataclasses.dataclass(frozen=True, order=True, slots=True, unsafe_hash=True)
class Marimba(Instrument):

    clefs: tuple[str, ...] = ("treble", "bass")
    context: str = "Staff"
    middle_c_sounding_pitch: _pitch.NamedPitch = _pitch.NamedPitch("C4")
    pitch_range: _pcollections.PitchRange = _pcollections.PitchRange("[F2, C7]")


@dataclasses.dataclass(frozen=True, order=True, slots=True, unsafe_hash=True)
class MezzoSopranoVoice(Instrument):

    clefs: tuple[str, ...] = ("treble",)
    context: str = "Staff"
    middle_c_sounding_pitch: _pitch.NamedPitch = _pitch.NamedPitch("C4")
    pitch_range: _pcollections.PitchRange = _pcollections.PitchRange("[A3, C6]")


@dataclasses.dataclass(frozen=True, order=True, slots=True, unsafe_hash=True)
class Oboe(Instrument):

    clefs: tuple[str, ...] = ("treble",)
    context: str = "Staff"
    middle_c_sounding_pitch: _pitch.NamedPitch = _pitch.NamedPitch("C4")
    pitch_range: _pcollections.PitchRange = _pcollections.PitchRange("[Bb3, A6]")


@dataclasses.dataclass(frozen=True, order=True, slots=True, unsafe_hash=True)
class Percussion(Instrument):

    clefs: tuple[str, ...] = ("percussion",)
    context: str = "Staff"
    middle_c_sounding_pitch: _pitch.NamedPitch = _pitch.NamedPitch("C4")
    pitch_range: _pcollections.PitchRange = _pcollections.PitchRange("[-inf, +inf]")

    known_percussion = tuple(
        sorted(
            set(
                [
                    "agogô",
                    "anvil",
                    "bass drum",
                    "bongo drums",
                    "cabasa",
                    "cajón",
                    "castanets",
                    "caxixi",
                    "claves",
                    "conga drums",
                    "cowbell",
                    "crotales",
                    "cuíca",
                    "djembe",
                    "finger cymbals",
                    "flexatone",
                    "frame drum",
                    "gong",
                    "güiro",
                    "hand-held stones",
                    "jawbone",
                    "maracas",
                    "ratchet",
                    "rattle",
                    "sand blocks",
                    "scraped slate",
                    "siren",
                    "slapstick",
                    "slide whistle",
                    "snare drum",
                    "sponges",
                    "suspended cymbal",
                    "steel drums",
                    "tam-tam",
                    "tambourine",
                    "temple blocks",
                    "thunder machine",
                    "thundersheet",
                    "toms",
                    "tubular bells",
                    "triangle",
                    "vibraslap",
                    "whistle",
                    "wind chime",
                    "wind machine",
                    "wood blocks",
                    "wood planks",
                ]
            )
        )
    )


@dataclasses.dataclass(frozen=True, order=True, slots=True, unsafe_hash=True)
class Piano(Instrument):

    clefs: tuple[str, ...] = ("treble", "bass")
    context: str = "StaffGroup"
    middle_c_sounding_pitch: _pitch.NamedPitch = _pitch.NamedPitch("C4")
    pitch_range: _pcollections.PitchRange = _pcollections.PitchRange("[A0, C8]")


@dataclasses.dataclass(frozen=True, order=True, slots=True, unsafe_hash=True)
class Piccolo(Instrument):

    clefs: tuple[str, ...] = ("treble",)
    context: str = "Staff"
    middle_c_sounding_pitch: _pitch.NamedPitch = _pitch.NamedPitch("C5")
    pitch_range: _pcollections.PitchRange = _pcollections.PitchRange("[D5, C8]")


@dataclasses.dataclass(frozen=True, order=True, slots=True, unsafe_hash=True)
class SopraninoSaxophone(Instrument):

    clefs: tuple[str, ...] = ("treble",)
    context: str = "Staff"
    middle_c_sounding_pitch: _pitch.NamedPitch = _pitch.NamedPitch("Eb4")
    pitch_range: _pcollections.PitchRange = _pcollections.PitchRange("[Db4, F#6]")


@dataclasses.dataclass(frozen=True, order=True, slots=True, unsafe_hash=True)
class SopranoSaxophone(Instrument):

    clefs: tuple[str, ...] = ("treble",)
    context: str = "Staff"
    middle_c_sounding_pitch: _pitch.NamedPitch = _pitch.NamedPitch("Bb3")
    pitch_range: _pcollections.PitchRange = _pcollections.PitchRange("[Ab3, E6]")


@dataclasses.dataclass(frozen=True, order=True, slots=True, unsafe_hash=True)
class SopranoVoice(Instrument):

    clefs: tuple[str, ...] = ("treble",)
    context: str = "Staff"
    middle_c_sounding_pitch: _pitch.NamedPitch = _pitch.NamedPitch("C4")
    pitch_range: _pcollections.PitchRange = _pcollections.PitchRange("[C4, E6]")


@dataclasses.dataclass(frozen=True, order=True, slots=True, unsafe_hash=True)
class TenorSaxophone(Instrument):

    clefs: tuple[str, ...] = ("treble",)
    context: str = "Staff"
    middle_c_sounding_pitch: _pitch.NamedPitch = _pitch.NamedPitch("Bb2")
    pitch_range: _pcollections.PitchRange = _pcollections.PitchRange("[Ab2, E5]")


@dataclasses.dataclass(frozen=True, order=True, slots=True, unsafe_hash=True)
class TenorTrombone(Instrument):

    clefs: tuple[str, ...] = ("tenor", "bass")
    context: str = "Staff"
    middle_c_sounding_pitch: _pitch.NamedPitch = _pitch.NamedPitch("C4")
    pitch_range: _pcollections.PitchRange = _pcollections.PitchRange("[E2, Eb5]")


@dataclasses.dataclass(frozen=True, order=True, slots=True, unsafe_hash=True)
class TenorVoice(Instrument):

    clefs: tuple[str, ...] = ("treble",)
    context: str = "Staff"
    middle_c_sounding_pitch: _pitch.NamedPitch = _pitch.NamedPitch("C4")
    pitch_range: _pcollections.PitchRange = _pcollections.PitchRange("[C3, D5]")


@dataclasses.dataclass(frozen=True, order=True, slots=True, unsafe_hash=True)
class Trumpet(Instrument):

    clefs: tuple[str, ...] = ("treble",)
    context: str = "Staff"
    middle_c_sounding_pitch: _pitch.NamedPitch = _pitch.NamedPitch("C4")
    pitch_range: _pcollections.PitchRange = _pcollections.PitchRange("[F#3, D6]")


@dataclasses.dataclass(frozen=True, order=True, slots=True, unsafe_hash=True)
class Tuba(Instrument):

    clefs: tuple[str, ...] = ("bass",)
    context: str = "Staff"
    middle_c_sounding_pitch: _pitch.NamedPitch = _pitch.NamedPitch("C4")
    pitch_range: _pcollections.PitchRange = _pcollections.PitchRange("[D1, F4]")


@dataclasses.dataclass(frozen=True, order=True, slots=True, unsafe_hash=True)
class Vibraphone(Instrument):

    clefs: tuple[str, ...] = ("treble",)
    context: str = "Staff"
    middle_c_sounding_pitch: _pitch.NamedPitch = _pitch.NamedPitch("C4")
    pitch_range: _pcollections.PitchRange = _pcollections.PitchRange("[F3, F6]")


@dataclasses.dataclass(frozen=True, order=True, slots=True, unsafe_hash=True)
class Viola(Instrument):

    clefs: tuple[str, ...] = ("alto", "treble")
    context: str = "Staff"
    middle_c_sounding_pitch: _pitch.NamedPitch = _pitch.NamedPitch("C4")
    pitch_range: _pcollections.PitchRange = _pcollections.PitchRange("[C3, D6]")
    tuning: Tuning = Tuning(
        tuple([_pitch.NamedPitch(_) for _ in "C3 G3 D4 A4".split()])
    )


@dataclasses.dataclass(frozen=True, order=True, slots=True, unsafe_hash=True)
class Violin(Instrument):

    clefs: tuple[str, ...] = ("treble",)
    context: str = "Staff"
    middle_c_sounding_pitch: _pitch.NamedPitch = _pitch.NamedPitch("C4")
    pitch_range: _pcollections.PitchRange = _pcollections.PitchRange("[G3, G7]")
    tuning: Tuning = Tuning(
        tuple([_pitch.NamedPitch(_) for _ in "G3 D4 A4 E5".split()])
    )


@dataclasses.dataclass(frozen=True, order=True, slots=True, unsafe_hash=True)
class Xylophone(Instrument):

    clefs: tuple[str, ...] = ("treble",)
    context: str = "Staff"
    middle_c_sounding_pitch: _pitch.NamedPitch = _pitch.NamedPitch("C5")
    pitch_range: _pcollections.PitchRange = _pcollections.PitchRange("[C4, C7]")



import dataclasses
import typing

from . import _indentlib


@dataclasses.dataclass(frozen=True, order=True, slots=True, unsafe_hash=True)
class Tag:
    """
    Tag.

    ..  container:: example

        >>> abjad.Tag("YELLOW")
        Tag(string='YELLOW')

        >>> abjad.Tag("YELLOW:RED")
        Tag(string='YELLOW:RED')

        Raises exception on duplicate words in tag:

        >>> abjad.Tag("YELLOW:RED:RED")
        Traceback (most recent call last):
        ...
        Exception: duplicate words in tag: 'YELLOW:RED:RED'

        Raises exception on multiple only-edition tags:

        >>> abjad.Tag("+SEGMENT:+PARTS")
        Traceback (most recent call last):
            ...
        Exception: at most one only-edition tag: ['+SEGMENT', '+PARTS'].

        Raises exception on mixed only-edition / not-edition tags:

        >>> abjad.Tag("+SEGMENT:-PARTS")
        Traceback (most recent call last):
        ...
        Exception: only-edition and not-edition forbidden in same tag:
        <BLANKLINE>
          ['+SEGMENT'] / ['-PARTS']

    """

    string: str = ""

    def __post_init__(self):
        assert isinstance(self.string, str), repr(self.string)
        self.words()

    def append(self, word: "Tag") -> "Tag":
        """
        Appends ``word`` to tag.

        ..  container:: example

            >>> abjad.Tag("-PARTS").append(abjad.Tag("DEFAULT_CLEF"))
            Tag(string='-PARTS:DEFAULT_CLEF')

        """
        if not bool(word.string):
            return Tag(self.string)
        assert isinstance(word, Tag), repr(word)
        words = []
        if self.string:
            words.append(self.string)
        if word.string in self.words():
            raise Exception(f"{word} duplicates {self}.")
        words.append(word.string)
        string = ":".join(words)
        return Tag(string)

    def editions(self) -> list["Tag"]:
        """
        Gets edition tags in tag.

        ..  container:: example

            >>> abjad.Tag("FOO").editions()
            []

            >>> abjad.Tag("+SEGMENT").only_edition()
            Tag(string='+SEGMENT')

            >>> abjad.Tag("+SEGMENT:FOO").only_edition()
            Tag(string='+SEGMENT')

            >>> abjad.Tag("-SEGMENT").editions()
            [Tag(string='-SEGMENT')]

            >>> abjad.Tag("-SEGMENT:FOO").editions()
            [Tag(string='-SEGMENT')]

            >>> abjad.Tag("-SEGMENT:-PARTS").editions()
            [Tag(string='-SEGMENT'), Tag(string='-PARTS')]

        """
        result = []
        for word in self.words():
            if word.startswith("+") or word.startswith("-"):
                result.append(Tag(word))
        return result

    def invert_edition_tags(self) -> "Tag":
        """
        Inverts edition tags in tag.

        ..  container:: example

            >>> abjad.Tag("FOO").invert_edition_tags()
            Tag(string='FOO')

            >>> abjad.Tag("FOO:-PARTS").invert_edition_tags()
            Tag(string='FOO:+PARTS')

            >>> abjad.Tag("FOO:+PARTS").invert_edition_tags()
            Tag(string='FOO:-PARTS')

        """
        words = []
        for word in self.words():
            if word.startswith("+"):
                word_ = "-" + word[1:]
            elif word.startswith("-"):
                word_ = "+" + word[1:]
            else:
                word_ = word
            words.append(word_)
        string = ":".join(words)
        tag = Tag(string)
        return tag

    def not_editions(self) -> list["Tag"]:
        """
        Gets not-edition tags in tag.

        ..  container:: example

            >>> abjad.Tag("FOO").not_editions()
            []

            >>> abjad.Tag("-SEGMENT").not_editions()
            [Tag(string='-SEGMENT')]

            >>> abjad.Tag("-SEGMENT:FOO").not_editions()
            [Tag(string='-SEGMENT')]

            >>> abjad.Tag("-SEGMENT:-PARTS").not_editions()
            [Tag(string='-SEGMENT'), Tag(string='-PARTS')]

        """
        result = []
        for word in self.words():
            if word.startswith("-"):
                result.append(Tag(word))
        return result

    def only_edition(self) -> typing.Optional["Tag"]:
        """
        Gets only-edition tag in tag.

        ..  container:: example

            >>> abjad.Tag("FOO").only_edition() is None
            True

            >>> abjad.Tag("+SEGMENT").only_edition()
            Tag(string='+SEGMENT')

            >>> abjad.Tag("+SEGMENT:FOO").only_edition()
            Tag(string='+SEGMENT')

        """
        for word in self.words():
            if word.startswith("+"):
                return Tag(word)
        else:
            return None

    def words(self) -> list[str]:
        """
        Gets words.

        ..  container:: example

            >>> abjad.Tag("-PARTS:DEFAULT_CLEF").words()
            ['-PARTS', 'DEFAULT_CLEF']

        """
        assert not self.string.startswith(":"), repr(self.string)
        words = self.string.split(":")
        assert isinstance(words, list), repr(words)
        words_ = []
        for word in words:
            if word in words_:
                raise Exception(f"duplicate words in tag: {self.string!r}")
            words_.append(word)
        only_edition_tags, not_edition_tags = [], []
        for word_ in words_:
            if word_.startswith("+"):
                only_edition_tags.append(word_)
            if word_.startswith("-"):
                not_edition_tags.append(word_)
        if 1 < len(only_edition_tags):
            raise Exception(f"at most one only-edition tag: {only_edition_tags!r}.")
        if only_edition_tags and not_edition_tags:
            message = "only-edition and not-edition forbidden in same tag:\n\n"
            message += f"  {only_edition_tags} / {not_edition_tags}"
            raise Exception(message)
        assert all(isinstance(_, str) for _ in words_), repr(words_)
        return words_


def _match_line(line, tag, current_tags):
    assert all(isinstance(_, Tag) for _ in current_tags), repr(current_tags)
    if tag in current_tags:
        return True
    if callable(tag):
        return tag(current_tags)
    assert isinstance(tag, Tag), repr(tag)
    return False


def activate(
    text: str, tag: Tag | typing.Callable, skipped: bool = False
) -> tuple[str, int] | tuple[str, int, int]:
    r"""
    Activates ``tag`` in ``text``.

    ..  container:: example

        Writes (deactivated) tag with ``"%@%"`` prefix into LilyPond input:

        >>> staff = abjad.Staff("c'4 d' e' f'")
        >>> markup = abjad.Markup(r"\markup { \with-color #red Allegro }")
        >>> abjad.attach(
        ...     markup,
        ...     staff[0],
        ...     deactivate=True,
        ...     tag=abjad.Tag("RED_MARKUP"),
        ... )

        >>> text = abjad.lilypond(staff, tags=True)
        >>> text = abjad.tag.left_shift_tags(text)
        >>> print(text)
        \new Staff
        {
            c'4
            %! RED_MARKUP
            %@% - \markup { \with-color #red Allegro }
            d'4
            e'4
            f'4
        }

        >>> abjad.show(staff) # doctest: +SKIP

        Activates tag:

        >>> text, count = abjad.activate(text, abjad.Tag("RED_MARKUP"))
        >>> print(text)
        \new Staff
        {
            c'4
            %! RED_MARKUP
            - \markup { \with-color #red Allegro } %@%
            d'4
            e'4
            f'4
        }

        >>> lines = [_.strip("\n") for _ in text.split("\n")]
        >>> string = "\n".join(lines)
        >>> lilypond_file = abjad.LilyPondFile([string])
        >>> abjad.show(lilypond_file) # doctest: +SKIP

        Deactivates tag again:

        >>> text, count = abjad.deactivate(text, abjad.Tag("RED_MARKUP"))
        >>> print(text)
        \new Staff
        {
            c'4
            %! RED_MARKUP
            %@% - \markup { \with-color #red Allegro }
            d'4
            e'4
            f'4
        }

        >>> lines = [_.strip("\n") for _ in text.split("\n")]
        >>> string = "\n".join(lines)
        >>> lilypond_file = abjad.LilyPondFile([string])
        >>> abjad.show(lilypond_file) # doctest: +SKIP

        Activates tag again:

        >>> text, count = abjad.activate(text, abjad.Tag("RED_MARKUP"))
        >>> print(text)
        \new Staff
        {
            c'4
            %! RED_MARKUP
            - \markup { \with-color #red Allegro } %@%
            d'4
            e'4
            f'4
        }

        >>> lines = [_.strip("\n") for _ in text.split("\n")]
        >>> string = "\n".join(lines)
        >>> lilypond_file = abjad.LilyPondFile([string])
        >>> abjad.show(lilypond_file) # doctest: +SKIP

    Tags can toggle indefinitely.

    Returns text, count pair.

    Count gives number of activated tags.
    """
    assert isinstance(text, str), repr(text)
    assert isinstance(tag, Tag) or callable(tag), repr(tag)
    lines: list[str] = []
    count, skipped_count = 0, 0
    treated_last_line = False
    found_already_active_on_last_line = False
    text_lines = text.split("\n")
    text_lines = [_ + "\n" for _ in text_lines[:-1]] + text_lines[-1:]
    lines = []
    current_tags = []
    for line in text_lines:
        if line.lstrip().startswith("%! "):
            lines.append(line)
            current_tag = Tag(line.strip()[3:])
            current_tags.append(current_tag)
            continue
        if not _match_line(line, tag, current_tags):
            lines.append(line)
            treated_last_line = False
            found_already_active_on_last_line = False
            current_tags = []
            continue
        first_nonwhitespace_index = len(line) - len(line.lstrip())
        index = first_nonwhitespace_index
        if line[index : index + 4] in ("%%% ", "%@% "):
            if "%@% " in line:
                line = line.replace("%@% ", "")
                suffix = " %@%"
            else:
                # TODO: replace with "" instead of "    "?
                line = line.replace("%%%", "   ")
                suffix = None
            assert line.endswith("\n"), repr(line)
            if suffix:
                line = line.strip("\n") + suffix + "\n"
            if not treated_last_line:
                count += 1
            treated_last_line = True
            found_already_active_on_last_line = False
        else:
            if not found_already_active_on_last_line:
                skipped_count += 1
            found_already_active_on_last_line = True
            treated_last_line = False
        lines.append(line)
        current_tags = []
    text = "".join(lines)
    if skipped is True:
        return text, count, skipped_count
    else:
        return text, count


def deactivate(
    text: str,
    tag: Tag | typing.Callable,
    prepend_empty_chord: bool = False,
    skipped: bool = False,
) -> tuple[str, int] | tuple[str, int, int]:
    r"""
    Deactivates ``tag`` in ``text``.

    ..  container:: example

        Writes (active) tag into LilyPond input:

        >>> staff = abjad.Staff("c'4 d' e' f'")
        >>> string = r"\markup { \with-color #red Allegro }"
        >>> markup = abjad.Markup(string)
        >>> abjad.attach(
        ...     markup,
        ...     staff[0],
        ...     tag=abjad.Tag("RED_MARKUP"),
        ... )

        >>> text = abjad.lilypond(staff, tags=True)
        >>> text = abjad.tag.left_shift_tags(text)
        >>> print(text)
        \new Staff
        {
            c'4
            %! RED_MARKUP
            - \markup { \with-color #red Allegro }
            d'4
            e'4
            f'4
        }

        >>> abjad.show(staff) # doctest: +SKIP

        Deactivates tag:

        >>> text = abjad.lilypond(staff, tags=True)
        >>> text, count = abjad.deactivate(text, abjad.Tag("RED_MARKUP"))
        >>> print(text)
        \new Staff
        {
            c'4
            %! RED_MARKUP
        %%% - \markup { \with-color #red Allegro }
            d'4
            e'4
            f'4
        }

        >>> lines = [_.strip("\n") for _ in text.split("\n")]
        >>> string = "\n".join(lines)
        >>> lilypond_file = abjad.LilyPondFile([string])
        >>> abjad.show(lilypond_file) # doctest: +SKIP

        Activates tag again:

        >>> text, count = abjad.activate(text, abjad.Tag("RED_MARKUP"))
        >>> print(text)
        \new Staff
        {
            c'4
            %! RED_MARKUP
            - \markup { \with-color #red Allegro }
            d'4
            e'4
            f'4
        }

        >>> lines = [_.strip("\n") for _ in text.split("\n")]
        >>> string = "\n".join(lines)
        >>> lilypond_file = abjad.LilyPondFile([string])
        >>> abjad.show(lilypond_file) # doctest: +SKIP

        Deactivates tag again:

        >>> text, count = abjad.deactivate(text, abjad.Tag("RED_MARKUP"))
        >>> print(text)
        \new Staff
        {
            c'4
            %! RED_MARKUP
        %%% - \markup { \with-color #red Allegro }
            d'4
            e'4
            f'4
        }

        >>> lines = [_.strip("\n") for _ in text.split("\n")]
        >>> string = "\n".join(lines)
        >>> lilypond_file = abjad.LilyPondFile([string])
        >>> abjad.show(lilypond_file) # doctest: +SKIP

    Tags can toggle indefinitely.
    """
    assert isinstance(text, str), repr(text)
    assert isinstance(tag, Tag) or callable(tag), repr(tag)
    count, skipped_count = 0, 0
    treated_last_line = False
    found_already_deactivated_on_last_line = False
    previous_line_was_tweak = False
    text_lines = text.split("\n")
    text_lines = [_ + "\n" for _ in text_lines[:-1]] + text_lines[-1:]
    lines, current_tags = [], []
    for line in text_lines:
        if line.lstrip().startswith("%! "):
            lines.append(line)
            current_tag = Tag(line.strip()[3:])
            current_tags.append(current_tag)
            continue
        if not _match_line(line, tag, current_tags):
            lines.append(line)
            treated_last_line = False
            found_already_deactivated_on_last_line = False
            current_tags = []
            continue
        start_column = len(line) - len(line.lstrip())
        if line[start_column] != "%":
            if " %@%" in line:
                prefix = "%@% "
                line = line.replace(" %@%", "")
            else:
                prefix = "%%% "
            if prepend_empty_chord and not previous_line_was_tweak:
                prefix += "<> "
            target = line[start_column - 4 : start_column]
            assert target == "    ", repr((line, target, start_column, tag))
            characters = list(line)
            characters[start_column - 4 : start_column] = list(prefix)
            line = "".join(characters)
            if not treated_last_line:
                count += 1
            treated_last_line = True
            found_already_deactivated_on_last_line = False
        else:
            if not found_already_deactivated_on_last_line:
                skipped_count += 1
            found_already_deactivated_on_last_line = True
            treated_last_line = False
        lines.append(line)
        previous_line_was_tweak = "tweak" in line
        current_tags = []
    text = "".join(lines)
    if skipped is True:
        return text, count, skipped_count
    else:
        return text, count


def double_tag(strings: list[str], tag_: Tag, deactivate: bool = False) -> list[str]:
    """
    Double tags ``strings``.
    """
    assert all(isinstance(_, str) for _ in strings), repr(strings)
    assert isinstance(tag_, Tag), repr(tag_)
    half_indent = 2 * " "
    tag_lines = []
    if tag_.string:
        tag_lines_ = tag_.string.split(":")
        tag_lines_ = [half_indent + "%! " + _ for _ in tag_lines_]
        tag_lines.extend(tag_lines_)
        tag_lines.sort()
    if deactivate is True:
        strings = ["%@% " + _ for _ in strings]
    result = []
    for string in strings:
        if string.strip().startswith("%!"):
            result.append(string)
            continue
        result.extend(tag_lines)
        result.append(string)
    return result


def left_shift_tags(text: str) -> str:
    """
    Left shifts tags in ``strings``.
    """
    strings = text.split("\n")
    strings_ = []
    for string in strings:
        if "%@% " not in string or "%!" not in string:
            strings_.append(string)
            continue
        if not string.startswith(4 * " "):
            strings_.append(string)
            continue
        string_ = string[4:]
        tag_start = string_.find("%!")
        strings__ = list(string_)
        strings__[tag_start:tag_start] = _indentlib.INDENT
        string_ = "".join(strings__)
        strings_.append(string_)
    text = "\n".join(strings_)
    return text


def remove_tags(string: str) -> str:
    """
    Removes all tags from ``string``.
    """
    lines = []
    for line in string.split("\n"):
        if "%!" not in line:
            lines.append(line)
            continue
        tag_start = line.find("%!")
        line = line[:tag_start]
        line = line.rstrip()
        if line:
            lines.append(line)
    string = "\n".join(lines)
    return string



import typing

from . import _getlib
from . import contributions as _contributions
from . import illustrators as _illustrators
from . import indicators as _indicators
from . import ratio as _ratio
from . import score as _score


class MetricModulation:
    r"""
    Metric modulation.

    ..  container:: example

        With notes:

        >>> metric_modulation = abjad.MetricModulation(
        ...     left_rhythm=abjad.Note("c'8."),
        ...     right_rhythm=abjad.Note("c'4."),
        ... )

        >>> lilypond_file = abjad.LilyPondFile(
        ...     [r'\include "abjad.ily"', metric_modulation]
        ... )
        >>> abjad.show(lilypond_file) # doctest: +SKIP

        ..  docs::

            >>> string = abjad.lilypond(metric_modulation)
            >>> print(string)
            \markup \abjad-metric-modulation #3 #1 #2 #1 #'(1 . 1)

    ..  container:: example

        With tuplets:

        >>> metric_modulation = abjad.MetricModulation(
        ...     left_rhythm=abjad.Tuplet((4, 5), "c'4"),
        ...     right_rhythm=abjad.Note("c'4"),
        ... )

        >>> lilypond_file = abjad.LilyPondFile(
        ...     [r'\include "abjad.ily"', metric_modulation]
        ... )
        >>> abjad.show(lilypond_file) # doctest: +SKIP

        ..  docs::

            >>> string = abjad.lilypond(metric_modulation)
            >>> print(string)
            \markup \abjad-metric-modulation-tuplet-lhs #2 #0 #4 #5 #2 #0 #'(1 . 1)

        >>> metric_modulation = abjad.MetricModulation(
        ...     left_rhythm=abjad.Note("c'4"),
        ...     right_rhythm=abjad.Tuplet((4, 5), "c'4"),
        ... )

        >>> lilypond_file = abjad.LilyPondFile(
        ...     [r'\include "abjad.ily"', metric_modulation]
        ... )
        >>> abjad.show(lilypond_file) # doctest: +SKIP

        ..  docs::

            >>> string = abjad.lilypond(metric_modulation)
            >>> print(string)
            \markup \abjad-metric-modulation-tuplet-rhs #2 #0 #2 #0 #4 #5 #'(1 . 1)

    ..  container:: example

        With tuplets again:

        >>> metric_modulation = abjad.MetricModulation(
        ...     left_rhythm=abjad.Note("c16."),
        ...     right_rhythm=abjad.Tuplet((2, 3), "c8"),
        ... )

        >>> lilypond_file = abjad.LilyPondFile(
        ...     [r'\include "abjad.ily"', metric_modulation]
        ... )
        >>> abjad.show(lilypond_file) # doctest: +SKIP

        ..  docs::

            >>> string = abjad.lilypond(metric_modulation)
            >>> print(string)
            \markup \abjad-metric-modulation-tuplet-rhs #4 #1 #3 #0 #2 #3 #'(1 . 1)

    ..  container:: example

        With ties:

        >>> notes = abjad.makers.make_notes([0], [(5, 16)])
        >>> metric_modulation = abjad.MetricModulation(
        ...     left_rhythm=abjad.Note("c'4"),
        ...     right_rhythm=notes,
        ... )

        >>> lilypond_file = abjad.LilyPondFile(
        ...     [r'\include "abjad.ily"', metric_modulation]
        ... )
        >>> abjad.show(lilypond_file) # doctest: +SKIP

        ..  docs::

            >>> string = abjad.lilypond(metric_modulation)
            >>> print(string)
            \markup { \score
                {
                    \context Score = "Score"
                    \with
                    {
                        \override SpacingSpanner.spacing-increment = 0.5
                        proportionalNotationDuration = ##f
                    }
                    <<
                        \context RhythmicStaff = "Rhythmic_Staff"
                        \with
                        {
                            \remove Time_signature_engraver
                            \remove Staff_symbol_engraver
                            \override Stem.direction = #up
                            \override Stem.length = 5
                            \override TupletBracket.bracket-visibility = ##t
                            \override TupletBracket.direction = #up
                            \override TupletBracket.minimum-length = 4
                            \override TupletBracket.padding = 1.25
                            \override TupletBracket.shorten-pair = #'(-1 . -1.5)
                            \override TupletBracket.springs-and-rods = #ly:spanner::set-spacing-rods
                            \override TupletNumber.font-size = 0
                            \override TupletNumber.text = #tuplet-number::calc-fraction-text
                            tupletFullLength = ##t
                        }
                        {
                            c'4
                        }
                    >>
                    \layout
                    {
                        indent = 0
                        ragged-right = ##t
                    }
                }
            =
            \hspace #-0.5
            \score
                {
                    \context Score = "Score"
                    \with
                    {
                        \override SpacingSpanner.spacing-increment = 0.5
                        proportionalNotationDuration = ##f
                    }
                    <<
                        \context RhythmicStaff = "Rhythmic_Staff"
                        \with
                        {
                            \remove Time_signature_engraver
                            \remove Staff_symbol_engraver
                            \override Stem.direction = #up
                            \override Stem.length = 5
                            \override TupletBracket.bracket-visibility = ##t
                            \override TupletBracket.direction = #up
                            \override TupletBracket.minimum-length = 4
                            \override TupletBracket.padding = 1.25
                            \override TupletBracket.shorten-pair = #'(-1 . -1.5)
                            \override TupletBracket.springs-and-rods = #ly:spanner::set-spacing-rods
                            \override TupletNumber.font-size = 0
                            \override TupletNumber.text = #tuplet-number::calc-fraction-text
                            tupletFullLength = ##t
                        }
                        {
                            c'4
                            ~
                            c'16
                        }
                    >>
                    \layout
                    {
                        indent = 0
                        ragged-right = ##t
                    }
                } }

    ..  container:: example

        With ties and tuplets:

        >>> notes = abjad.makers.make_notes([0], [(5, 16)])
        >>> tuplet = abjad.Tuplet((2, 3), notes)
        >>> metric_modulation = abjad.MetricModulation(
        ...     left_rhythm=abjad.Note("c'4"),
        ...     right_rhythm=tuplet,
        ... )

        >>> lilypond_file = abjad.LilyPondFile(
        ...     [r'\include "abjad.ily"', metric_modulation]
        ... )
        >>> abjad.show(lilypond_file) # doctest: +SKIP

        ..  docs::

            >>> string = abjad.lilypond(metric_modulation)
            >>> print(string)
            \markup { \score
                {
                    \context Score = "Score"
                    \with
                    {
                        \override SpacingSpanner.spacing-increment = 0.5
                        proportionalNotationDuration = ##f
                    }
                    <<
                        \context RhythmicStaff = "Rhythmic_Staff"
                        \with
                        {
                            \remove Time_signature_engraver
                            \remove Staff_symbol_engraver
                            \override Stem.direction = #up
                            \override Stem.length = 5
                            \override TupletBracket.bracket-visibility = ##t
                            \override TupletBracket.direction = #up
                            \override TupletBracket.minimum-length = 4
                            \override TupletBracket.padding = 1.25
                            \override TupletBracket.shorten-pair = #'(-1 . -1.5)
                            \override TupletBracket.springs-and-rods = #ly:spanner::set-spacing-rods
                            \override TupletNumber.font-size = 0
                            \override TupletNumber.text = #tuplet-number::calc-fraction-text
                            tupletFullLength = ##t
                        }
                        {
                            c'4
                        }
                    >>
                    \layout
                    {
                        indent = 0
                        ragged-right = ##t
                    }
                }
            =
            \hspace #-0.5
            \score
                {
                    \context Score = "Score"
                    \with
                    {
                        \override SpacingSpanner.spacing-increment = 0.5
                        proportionalNotationDuration = ##f
                    }
                    <<
                        \context RhythmicStaff = "Rhythmic_Staff"
                        \with
                        {
                            \remove Time_signature_engraver
                            \remove Staff_symbol_engraver
                            \override Stem.direction = #up
                            \override Stem.length = 5
                            \override TupletBracket.bracket-visibility = ##t
                            \override TupletBracket.direction = #up
                            \override TupletBracket.minimum-length = 4
                            \override TupletBracket.padding = 1.25
                            \override TupletBracket.shorten-pair = #'(-1 . -1.5)
                            \override TupletBracket.springs-and-rods = #ly:spanner::set-spacing-rods
                            \override TupletNumber.font-size = 0
                            \override TupletNumber.text = #tuplet-number::calc-fraction-text
                            tupletFullLength = ##t
                        }
                        {
                            \tweak edge-height #'(0.7 . 0)
                            \times 2/3
                            {
                                c'4
                                ~
                                c'16
                            }
                        }
                    >>
                    \layout
                    {
                        indent = 0
                        ragged-right = ##t
                    }
                } }


    ..  container:: example

        Attach metric modulations to generate score output:

        >>> staff = abjad.Staff("c'4 d'4 e'4 f'4 e'4 d'4")
        >>> abjad.attach(abjad.TimeSignature((3, 4)), staff[0])
        >>> score = abjad.Score([staff])

        >>> metric_modulation = abjad.MetricModulation(
        ...     left_rhythm=abjad.Note("c4"),
        ...     right_rhythm=abjad.Note("c8."),
        ... )
        >>> abjad.attach(metric_modulation, staff[3], direction=abjad.UP)
        >>> abjad.override(staff).TextScript.staff_padding = 2.5

        >>> lilypond_file = abjad.LilyPondFile([r'\include "abjad.ily"', score])
        >>> abjad.show(lilypond_file) # doctest: +SKIP

        ..  docs::

            >>> string = abjad.lilypond(score)
            >>> print(string)
            \new Score
            <<
                \new Staff
                \with
                {
                    \override TextScript.staff-padding = 2.5
                }
                {
                    \time 3/4
                    c'4
                    d'4
                    e'4
                    f'4
                    ^ \markup \abjad-metric-modulation #2 #0 #3 #1 #'(1 . 1)
                    e'4
                    d'4
                }
            >>

    """

    __slots__ = (
        "_hide",
        "_left_markup",
        "_left_rhythm",
        "_right_markup",
        "_right_rhythm",
        "_scale",
    )

    directed: typing.ClassVar[bool] = True

    def __init__(
        self,
        left_rhythm,
        right_rhythm,
        *,
        hide: bool = False,
        left_markup: _indicators.Markup = None,
        right_markup: _indicators.Markup = None,
        scale: tuple[int | float, int | float] = (1, 1),
    ) -> None:
        self._hide = bool(hide)
        left_rhythm = self._initialize_rhythm(left_rhythm)
        self._left_rhythm = left_rhythm
        right_rhythm = self._initialize_rhythm(right_rhythm)
        self._right_rhythm = right_rhythm
        self._right_rhythm = right_rhythm
        if left_markup is not None:
            assert isinstance(left_markup, _indicators.Markup)
        self._left_markup = left_markup
        if right_markup is not None:
            assert isinstance(right_markup, _indicators.Markup)
        self._right_markup = right_markup
        assert isinstance(scale, tuple), repr(scale)
        self._scale = scale

    def __eq__(self, argument) -> bool:
        """
        Is true when ``argument`` is another metric modulation with the same
        ratio as this metric modulation.

        ..  container:: example

            >>> metric_modulation_1 = abjad.MetricModulation(
            ...     left_rhythm=abjad.Note("c'4"),
            ...     right_rhythm=abjad.Note("c'4."),
            ... )
            >>> metric_modulation_2 = abjad.MetricModulation(
            ...     left_rhythm=abjad.Tuplet((2, 3), [abjad.Note("c'4")]),
            ...     right_rhythm=abjad.Note("c'4"),
            ... )
            >>> notes = abjad.makers.make_notes([0], [(5, 16)])
            >>> metric_modulation_3 = abjad.MetricModulation(
            ...     left_rhythm=abjad.Note("c'4"),
            ...     right_rhythm=notes,
            ... )

            >>> metric_modulation_1.ratio
            Ratio(numbers=(2, 3))
            >>> metric_modulation_2.ratio
            Ratio(numbers=(2, 3))
            >>> metric_modulation_3.ratio
            Ratio(numbers=(4, 5))

            >>> metric_modulation_1 == metric_modulation_1
            True
            >>> metric_modulation_1 == metric_modulation_2
            True
            >>> metric_modulation_1 == metric_modulation_3
            False

            >>> metric_modulation_2 == metric_modulation_1
            True
            >>> metric_modulation_2 == metric_modulation_2
            True
            >>> metric_modulation_2 == metric_modulation_3
            False

            >>> metric_modulation_3 == metric_modulation_1
            False
            >>> metric_modulation_3 == metric_modulation_2
            False
            >>> metric_modulation_3 == metric_modulation_3
            True

        """
        # custom definition because input rhythms don't compare:
        if isinstance(argument, type(self)):
            if self.ratio == argument.ratio:
                return True
        return False

    def __hash__(self) -> int:
        """
        Hashes metric modulation.
        """
        return hash(self.__class__.__name__ + str(self))

    def __repr__(self) -> str:
        """
        Gets repr.
        """
        return f"{type(self).__name__}(left_rhythm={self.left_rhythm!r}, right_rhythm={self.right_rhythm!r})"

    ### PRIVATE METHODS ###

    def _get_lilypond_command_string(self):
        if self._note_to_note():
            arguments = self._get_markup_arguments()
            left_exponent, left_dots, right_exponent, right_dots = arguments
            string = r"\abjad-metric-modulation"
            string += f" #{left_exponent} #{left_dots}"
            string += f" #{right_exponent} #{right_dots}"
        elif self._rhs_tuplet():
            arguments = self._get_markup_arguments()
            note_exponent, note_dots = arguments[:2]
            tuplet_exponent, tuplet_dots, tuplet_n, tuplet_d = arguments[2:]
            string = r"\abjad-metric-modulation-tuplet-rhs"
            string += f" #{note_exponent} #{note_dots}"
            string += f" #{tuplet_exponent} #{tuplet_dots}"
            string += f" #{tuplet_n} #{tuplet_d}"
        elif self._lhs_tuplet():
            arguments = self._get_markup_arguments()
            tuplet_exponent, tuplet_dots, tuplet_n, tuplet_d = arguments[:4]
            note_exponent, note_dots = arguments[4:]
            string = r"\abjad-metric-modulation-tuplet-lhs"
            string += f" #{tuplet_exponent} #{tuplet_dots}"
            string += f" #{tuplet_n} #{tuplet_d}"
            string += f" #{note_exponent} #{note_dots}"
        else:
            return None
        string += f" #'({self.scale[0]} . {self.scale[1]})"
        return string

    def _get_lilypond_format(self):
        markup = self._get_markup()
        return markup.string

    def _get_contributions(self, *, component=None, wrapper=None):
        contributions = _contributions.ContributionsBySite()
        if not self.hide:
            markup = self._get_markup()
            contributions = markup._get_contributions(
                component=component, wrapper=wrapper
            )
        return contributions

    def _get_markup(self):
        string = self._get_lilypond_command_string()
        if string is not None:
            markup = _indicators.Markup(rf"\markup {string}")
            return markup
        strings = []
        string = _illustrators.selection_to_score_markup_string(self.left_rhythm)
        strings.extend(string.split("\n"))
        strings.append("=")
        strings.append(r"\hspace #-0.5")
        string = _illustrators.selection_to_score_markup_string(self.right_rhythm)
        strings.extend(string.split("\n"))
        string = "\n".join(strings)
        string = rf"\markup {{ {string} }}"
        markup = _indicators.Markup(string)
        return markup

    def _get_markup_arguments(self):
        if self._note_to_note():
            left_exponent = self.left_rhythm[0].written_duration.exponent
            left_dots = self.left_rhythm[0].written_duration.dot_count
            right_exponent = self.right_rhythm[0].written_duration.exponent
            right_dots = self.right_rhythm[0].written_duration.dot_count
            return (left_exponent, left_dots, right_exponent, right_dots)
        elif self._lhs_tuplet():
            tuplet_exponent = self.left_rhythm[0][0].written_duration.exponent
            tuplet_dots = self.left_rhythm[0][0].written_duration.dot_count
            tuplet_n = self.left_rhythm[0].multiplier.numerator
            tuplet_d = self.left_rhythm[0].multiplier.denominator
            note_exponent = self.right_rhythm[0].written_duration.exponent
            note_dots = self.right_rhythm[0].written_duration.dot_count
            return (
                tuplet_exponent,
                tuplet_dots,
                tuplet_n,
                tuplet_d,
                note_exponent,
                note_dots,
            )
        elif self._rhs_tuplet():
            note_exponent = self.left_rhythm[0].written_duration.exponent
            note_dots = self.left_rhythm[0].written_duration.dot_count
            tuplet_exponent = self.right_rhythm[0][0].written_duration.exponent
            tuplet_dots = self.right_rhythm[0][0].written_duration.dot_count
            tuplet_n = self.right_rhythm[0].multiplier.numerator
            tuplet_d = self.right_rhythm[0].multiplier.denominator
            return (
                note_exponent,
                note_dots,
                tuplet_exponent,
                tuplet_dots,
                tuplet_n,
                tuplet_d,
            )
        else:
            raise Exception("implement tied note values.")

    def _initialize_rhythm(self, rhythm):
        if isinstance(rhythm, _score.Component):
            selection = [rhythm]
        else:
            selection = rhythm
        return selection

    def _lhs_tuplet(self):
        if (
            isinstance(self.left_rhythm[0], _score.Tuplet)
            and len(self.left_rhythm[0]) == 1
            and isinstance(self.right_rhythm[0], _score.Note)
            and len(self.right_rhythm) == 1
        ):
            return True
        return False

    def _note_to_note(self):
        if (
            isinstance(self.left_rhythm[0], _score.Note)
            and len(self.left_rhythm) == 1
            and isinstance(self.right_rhythm[0], _score.Note)
            and len(self.right_rhythm) == 1
        ):
            return True
        return False

    def _rhs_tuplet(self):
        if (
            isinstance(self.left_rhythm[0], _score.Note)
            and len(self.left_rhythm) == 1
            and isinstance(self.right_rhythm[0], _score.Tuplet)
            and len(self.right_rhythm[0]) == 1
        ):
            return True
        return False

    ### PUBLIC PROPERTIES ###

    @property
    def hide(self) -> bool:
        """
        Is true when metric modulation generates no LilyPond output.
        """
        return self._hide

    @property
    def left_markup(self) -> _indicators.Markup | None:
        """
        Gets left markup of metric modulation.

        ..  container:: example

            >>> metric_modulation = abjad.MetricModulation(
            ...     left_rhythm=abjad.Note("c'4"),
            ...     right_rhythm=abjad.Note("c'4."),
            ... )
            >>> metric_modulation.left_markup

        """
        return self._left_markup

    @property
    def left_rhythm(self):
        """
        Gets left rhythm of metric modulation.

        ..  container:: example

            >>> metric_modulation = abjad.MetricModulation(
            ...     left_rhythm=abjad.Note("c'4"),
            ...     right_rhythm=abjad.Note("c'4."),
            ... )
            >>> metric_modulation.left_rhythm
            [Note("c'4")]

        Returns selection.
        """
        return self._left_rhythm

    @property
    def ratio(self) -> _ratio.Ratio:
        """
        Gets ratio of metric modulation.

        ..  container:: example

            >>> metric_modulation = abjad.MetricModulation(
            ...     left_rhythm=abjad.Tuplet((2, 3), [abjad.Note("c'4")]),
            ...     right_rhythm=abjad.Note("c'4"),
            ... )
            >>> metric_modulation.ratio
            Ratio(numbers=(2, 3))

        """
        left_duration = _getlib._get_duration(self.left_rhythm)
        right_duration = _getlib._get_duration(self.right_rhythm)
        duration = left_duration / right_duration
        ratio = _ratio.Ratio(duration.pair)
        return ratio

    @property
    def right_markup(self) -> _indicators.Markup | None:
        r"""Gets right markup of metric modulation.

        ..  container:: example

            >>> metric_modulation = abjad.MetricModulation(
            ...     left_rhythm=abjad.Note("c'4"),
            ...     right_rhythm=abjad.Note("c'4."),
            ... )
            >>> metric_modulation.right_markup

        """
        return self._right_markup

    @property
    def right_rhythm(self):
        """
        Gets right tempo of metric modulation.

        ..  container:: example

            >>> metric_modulation = abjad.MetricModulation(
            ...     left_rhythm=abjad.Note("c'4"),
            ...     right_rhythm=abjad.Note("c'4."),
            ... )
            >>> metric_modulation.right_rhythm
            [Note("c'4.")]

        """
        return self._right_rhythm

    @property
    def scale(self) -> tuple[int | float, int | float]:
        r"""
        Gets scale of output markup.

        ..  container:: example

            >>> staff = abjad.Staff("c'4 d'4 e'4 f'4 e'4 d'4")
            >>> abjad.attach(abjad.TimeSignature((3, 4)), staff[0])
            >>> score = abjad.Score([staff])

            >>> metric_modulation = abjad.MetricModulation(
            ...     left_rhythm=abjad.Note("c4"),
            ...     right_rhythm=abjad.Note("c8."),
            ...     scale=(0.5, 0.5),
            ... )
            >>> abjad.attach(metric_modulation, staff[3], direction=abjad.UP)
            >>> abjad.override(staff).TextScript.staff_padding = 2.5

            >>> lilypond_file = abjad.LilyPondFile([r'\include "abjad.ily"', score])
            >>> abjad.show(lilypond_file) # doctest: +SKIP

            ..  docs::

                >>> string = abjad.lilypond(score)
                >>> print(string)
                \new Score
                <<
                    \new Staff
                    \with
                    {
                        \override TextScript.staff-padding = 2.5
                    }
                    {
                        \time 3/4
                        c'4
                        d'4
                        e'4
                        f'4
                        ^ \markup \abjad-metric-modulation #2 #0 #3 #1 #'(0.5 . 0.5)
                        e'4
                        d'4
                    }
                >>

        """
        return self._scale

    @property
    def tweaks(self) -> None:
        """
        Are not implemented on metric modulation.
        """
        pass



import enum
import typing

from . import _getlib
from . import duration as _duration
from . import indicators as _indicators
from . import instruments as _instruments
from . import iterate as _iterate
from . import iterpitches as _iterpitches
from . import parentage as _parentage
from . import score as _score
from . import sequence as _sequence
from . import tag as _tag


def _aggregate_context_wrappers(argument):
    """
    Special_Voice may contain other instances of Special_Voice.
    This currently happens with OnBeatGraceContainer.
    This method aggregates all Special_Voice wrappers for checks.
    """
    name_to_wrappers: dict = {}
    for context in _iterate.components(argument, _score.Context):
        if context.name not in name_to_wrappers:
            name_to_wrappers[context.name] = []
        wrappers = context._dependent_wrappers[:]
        name_to_wrappers[context.name].extend(wrappers)
    return name_to_wrappers


def check_beamed_long_notes(argument) -> tuple[list, int]:
    r"""
    Checks beamed long notes.

    ..  container:: example

        Beamed quarter notes are not wellformed:

        >>> voice = abjad.Voice("c'4 d'4 e'4 f'4")
        >>> abjad.attach(abjad.StartBeam(), voice[0])
        >>> abjad.attach(abjad.StopBeam(), voice[1])
        >>> abjad.show(voice) # doctest: +SKIP

        ..  docs::

            >>> string = abjad.lilypond(voice)
            >>> print(string)
            \new Voice
            {
                c'4
                [
                d'4
                ]
                e'4
                f'4
            }

        >>> count, string = abjad.wf.tabulate_wellformedness(voice)
        >>> print(string)
        2 /    4 beamed long notes
        0 /    5 duplicate ids
        0 /    1 empty containers
        0 /    5 missing parents
        0 /    4 notes on wrong clef
        0 /    4 out of range pitches
        0 /    0 overlapping text spanners
        0 /    0 unmatched stop text spans
        0 /    0 unterminated hairpins
        0 /    0 unterminated text spanners

        Beamed eighth notes are wellformed:

        >>> voice = abjad.Voice("c'8 d'8 e'8 f'8")
        >>> abjad.attach(abjad.StartBeam(), voice[0])
        >>> abjad.attach(abjad.StopBeam(), voice[1])
        >>> abjad.show(voice) # doctest: +SKIP

        ..  docs::

            >>> string = abjad.lilypond(voice)
            >>> print(string)
            \new Voice
            {
                c'8
                [
                d'8
                ]
                e'8
                f'8
            }

        >>> count, string = abjad.wf.tabulate_wellformedness(voice)
        >>> print(string)
        0 /    4 beamed long notes
        0 /    5 duplicate ids
        0 /    1 empty containers
        0 /    5 missing parents
        0 /    4 notes on wrong clef
        0 /    4 out of range pitches
        0 /    0 overlapping text spanners
        0 /    0 unmatched stop text spans
        0 /    0 unterminated hairpins
        0 /    0 unterminated text spanners

    ..  container:: example

        REGRESSION. Matching start- and stop-beam indicators work
        correctly:

        >>> voice = abjad.Voice("c'8 d'8 e'4 f'2")
        >>> abjad.attach(abjad.StartBeam(), voice[0])
        >>> abjad.attach(abjad.StopBeam(), voice[1])
        >>> abjad.show(voice) # doctest: +SKIP

        ..  docs::

            >>> string = abjad.lilypond(voice)
            >>> print(string)
            \new Voice
            {
                c'8
                [
                d'8
                ]
                e'4
                f'2
            }

        >>> count, string = abjad.wf.tabulate_wellformedness(voice)
        >>> print(string)
        0 /    4 beamed long notes
        0 /    5 duplicate ids
        0 /    1 empty containers
        0 /    5 missing parents
        0 /    4 notes on wrong clef
        0 /    4 out of range pitches
        0 /    0 overlapping text spanners
        0 /    0 unmatched stop text spans
        0 /    0 unterminated hairpins
        0 /    0 unterminated text spanners

    The examples above feature Abjad voice containers because beams are
    voice-persistent.
    """
    violators, total = [], 0
    for leaf in _iterate.leaves(argument):
        total += 1
        if leaf.written_duration < _duration.Duration((1, 4)):
            continue
        start_wrapper = _getlib._get_effective(
            leaf, _indicators.StartBeam, unwrap=False
        )
        if start_wrapper is None:
            continue
        stop_wrapper = _getlib._get_effective(leaf, _indicators.StopBeam, unwrap=False)
        if stop_wrapper is None:
            violators.append(leaf)
            continue
        if stop_wrapper.leaked_start_offset < start_wrapper.leaked_start_offset:
            violators.append(leaf)
            continue
        leaf_start_offset = leaf._get_timespan().start_offset
        if stop_wrapper.leaked_start_offset == leaf_start_offset:
            violators.append(leaf)
    return violators, total


def check_duplicate_ids(argument) -> tuple[list, int]:
    """
    Checks duplicate IDs.
    """
    violators = []
    components = _iterate.components(argument)
    total_ids = [id(_) for _ in components]
    unique_ids = _sequence.remove_repeats(total_ids)
    if len(unique_ids) < len(total_ids):
        for current_id in unique_ids:
            if 1 < total_ids.count(current_id):
                violators.extend([_ for _ in components if id(_) == current_id])
    return violators, len(total_ids)


def check_empty_containers(argument) -> tuple[list, int]:
    r"""
    Checks empty containers.

    ..  container:: example

        >>> staff = abjad.Staff("c'4 d' e' f'")
        >>> staff.append(abjad.Container())

        >>> string = abjad.lilypond(staff)
        >>> print(string)
        \new Staff
        {
            c'4
            d'4
            e'4
            f'4
            {
            }
        }

        >>> violators, total = abjad.wf.check_empty_containers(staff)
        >>> violators
        [Container()]

    """
    violators, containers = [], set()
    for container in _iterate.components(argument, _score.Container):
        containers.add(container)
        if len(container) == 0:
            violators.append(container)
    return violators, len(containers)


def check_missing_parents(argument) -> tuple[list, int]:
    """
    Checks missing parents.
    """
    violators, total = [], set()
    components = _iterate.components(argument)
    for i, component in enumerate(components):
        total.add(component)
        if 0 < i:
            # TODO: figure out why "if component._parent is None" doesn't work
            if _parentage.Parentage(component).parent is None:
                violators.append(component)
    return violators, len(total)


def check_notes_on_wrong_clef(argument) -> tuple[list, int]:
    r"""
    Checks notes and chords on wrong clef.

    ..  container:: example

        >>> staff = abjad.Staff("c'8 d'8 e'8 f'8")
        >>> clef = abjad.Clef('alto')
        >>> abjad.attach(clef, staff[0])
        >>> violin = abjad.Violin()
        >>> abjad.attach(violin, staff[0])
        >>> abjad.show(staff) # doctest: +SKIP

        ..  docs::

            >>> string = abjad.lilypond(staff)
            >>> print(string)
            \new Staff
            {
                \clef "alto"
                c'8
                d'8
                e'8
                f'8
            }

        >>> count, string = abjad.wf.tabulate_wellformedness(staff)
        >>> print(string)
        0 /    4 beamed long notes
        0 /    5 duplicate ids
        0 /    1 empty containers
        0 /    5 missing parents
        4 /    4 notes on wrong clef
        0 /    4 out of range pitches
        0 /    0 overlapping text spanners
        0 /    0 unmatched stop text spans
        0 /    0 unterminated hairpins
        0 /    0 unterminated text spanners

    ..  container:: example

        Always allows percussion clef:

        >>> staff = abjad.Staff("c'8 d'8 e'8 f'8")
        >>> clef = abjad.Clef('percussion')
        >>> abjad.attach(clef, staff[0])
        >>> violin = abjad.Violin()
        >>> abjad.attach(violin, staff[0])
        >>> abjad.show(staff) # doctest: +SKIP

        ..  docs::

            >>> string = abjad.lilypond(staff)
            >>> print(string)
            \new Staff
            {
                \clef "percussion"
                c'8
                d'8
                e'8
                f'8
            }

        >>> count, string = abjad.wf.tabulate_wellformedness(staff)
        >>> print(string)
        0 /    4 beamed long notes
        0 /    5 duplicate ids
        0 /    1 empty containers
        0 /    5 missing parents
        0 /    4 notes on wrong clef
        0 /    4 out of range pitches
        0 /    0 overlapping text spanners
        0 /    0 unmatched stop text spans
        0 /    0 unterminated hairpins
        0 /    0 unterminated text spanners

    """
    violators, total = [], set()
    for leaf in _iterate.leaves(argument):
        total.add(leaf)
        instrument = _getlib._get_effective(leaf, _instruments.Instrument)
        if instrument is None:
            continue
        effective_clef = _getlib._get_effective(leaf, _indicators.Clef)
        if effective_clef is None:
            continue
        clefs = []
        for clef in instrument.clefs:
            if isinstance(clef, str):
                clef = _indicators.Clef(clef)
            assert isinstance(clef, _indicators.Clef), repr(clef)
            clefs.append(clef)
        clefs.append(_indicators.Clef("percussion"))
        if effective_clef not in clefs:
            violators.append(leaf)
    return violators, len(total)


def check_out_of_range_pitches(
    argument, *, allow_indicators: typing.Sequence[str | enum.Enum] = ()
) -> tuple[list, int]:
    r"""
    Checks out-of-range notes.

    ..  container:: example

        Out of range:

        >>> staff = abjad.Staff("c'8 r8 <d fs>8 r8")
        >>> violin = abjad.Violin()
        >>> abjad.attach(violin, staff[0])
        >>> abjad.show(staff) # doctest: +SKIP

        ..  docs::

            >>> string = abjad.lilypond(staff)
            >>> print(string)
            \new Staff
            {
                c'8
                r8
                <d fs>8
                r8
            }

        >>> count, string = abjad.wf.tabulate_wellformedness(staff)
        >>> print(string)
        0 /    4 beamed long notes
        0 /    5 duplicate ids
        0 /    1 empty containers
        0 /    5 missing parents
        0 /    4 notes on wrong clef
        1 /    2 out of range pitches
        0 /    0 overlapping text spanners
        0 /    0 unmatched stop text spans
        0 /    0 unterminated hairpins
        0 /    0 unterminated text spanners

    ..  container:: example

        Using ``allow_indicators``:

        >>> staff = abjad.Staff("c'8 r8 <d fs>8 r8")
        >>> violin = abjad.Violin()
        >>> abjad.attach(violin, staff[0])
        >>> abjad.attach("ALLOW_OUT_OF_RANGE", staff[2])
        >>> abjad.show(staff) # doctest: +SKIP

        ..  docs::

            >>> string = abjad.lilypond(staff)
            >>> print(string)
            \new Staff
            {
                c'8
                r8
                <d fs>8
                r8
            }

        Does not check for indicator:

        >>> violators, total = abjad.wf.check_out_of_range_pitches(staff)
        >>> violators
        [Chord('<d fs>8')]

        Does check for indicator:

        >>> violators, total = abjad.wf.check_out_of_range_pitches(
        ...     staff, allow_indicators=["ALLOW_OUT_OF_RANGE"]
        ... )
        >>> violators
        []

    """
    violators, total = [], set()
    for leaf in _iterate.leaves(argument, pitched=True):
        total.add(leaf)
        ok = False
        for indicator in allow_indicators or ():
            if leaf._has_indicator(indicator):
                ok = True
        if ok is True:
            continue
        if "unpitched" in argument._get_indicators(str):
            continue
        instrument = _getlib._get_effective(leaf, _instruments.Instrument)
        if instrument is None:
            continue
        if not _iterpitches.sounding_pitches_are_in_range(leaf, instrument.pitch_range):
            violators.append(leaf)
    return violators, len(total)


def check_overlapping_text_spanners(argument) -> tuple[list, int]:
    r"""
    Checks overlapping text spanners.

    ..  container:: example

        Overlapping text spanners are not wellformed:

        >>> voice = abjad.Voice("c'4 c'4 c'4 c'4")
        >>> abjad.text_spanner(voice)
        >>> abjad.text_spanner(voice[1:3])
        >>> string = abjad.lilypond(voice)
        >>> print(string)
        \new Voice
        {
            c'4
            \startTextSpan
            c'4
            \startTextSpan
            c'4
            \stopTextSpan
            c'4
            \stopTextSpan
        }

        >>> count, string = abjad.wf.tabulate_wellformedness(voice)
        >>> print(string)
        0 /    4 beamed long notes
        0 /    5 duplicate ids
        0 /    1 empty containers
        0 /    5 missing parents
        0 /    4 notes on wrong clef
        0 /    4 out of range pitches
        1 /    2 overlapping text spanners
        0 /    2 unmatched stop text spans
        0 /    0 unterminated hairpins
        0 /    2 unterminated text spanners

    ..  container:: example

        Overlapping text spanners with different IDs are wellformed:

        >>> voice = abjad.Voice("c'4 c'4 c'4 c'4")
        >>> abjad.text_spanner(voice)
        >>> command = r'\startTextSpanOne'
        >>> start_text_span = abjad.StartTextSpan(command=command)
        >>> abjad.attach(start_text_span, voice[1])
        >>> command = r'\stopTextSpanOne'
        >>> stop_text_span = abjad.StopTextSpan(command=command)
        >>> abjad.attach(stop_text_span, voice[2])
        >>> string = abjad.lilypond(voice)
        >>> print(string)
        \new Voice
        {
            c'4
            \startTextSpan
            c'4
            \startTextSpanOne
            c'4
            \stopTextSpanOne
            c'4
            \stopTextSpan
        }

        >>> count, string = abjad.wf.tabulate_wellformedness(voice)
        >>> print(string)
        0 /    4 beamed long notes
        0 /    5 duplicate ids
        0 /    1 empty containers
        0 /    5 missing parents
        0 /    4 notes on wrong clef
        0 /    4 out of range pitches
        0 /    2 overlapping text spanners
        0 /    2 unmatched stop text spans
        0 /    0 unterminated hairpins
        0 /    2 unterminated text spanners

    ..  container:: example

        Enchained text spanners do not overlap (and are wellformed):

        >>> voice = abjad.Voice("c'4 c'4 c'4 c'4")
        >>> abjad.text_spanner(voice[:3])
        >>> abjad.text_spanner(voice[-2:])
        >>> string = abjad.lilypond(voice)
        >>> print(string)
        \new Voice
        {
            c'4
            \startTextSpan
            c'4
            c'4
            \stopTextSpan
            \startTextSpan
            c'4
            \stopTextSpan
        }

        >>> count, string = abjad.wf.tabulate_wellformedness(voice)
        >>> print(string)
        0 /    4 beamed long notes
        0 /    5 duplicate ids
        0 /    1 empty containers
        0 /    5 missing parents
        0 /    4 notes on wrong clef
        0 /    4 out of range pitches
        0 /    2 overlapping text spanners
        0 /    2 unmatched stop text spans
        0 /    0 unterminated hairpins
        0 /    2 unterminated text spanners

    ..  container:: example

        REGRESSION. Matching start- and stop-text-spans on a single leaf do
        not overlap (and are wellformed) iff stop-text-span leaks to the
        right:

        >>> voice = abjad.Voice("c'2 d'2 e'2 f'2")
        >>> abjad.attach(abjad.StartTextSpan(), voice[0])
        >>> stop_text_span = abjad.StopTextSpan(leak=True)
        >>> abjad.attach(stop_text_span, voice[0])
        >>> abjad.attach(abjad.StartTextSpan(), voice[2])
        >>> stop_text_span = abjad.StopTextSpan()
        >>> abjad.attach(stop_text_span, voice[3])
        >>> abjad.show(voice) # doctest: +SKIP

        >>> string = abjad.lilypond(voice)
        >>> print(string)
        \new Voice
        {
            c'2
            \startTextSpan
            <> \stopTextSpan
            d'2
            e'2
            \startTextSpan
            f'2
            \stopTextSpan
        }

        >>> count, string = abjad.wf.tabulate_wellformedness(voice)
        >>> print(string)
        0 /    4 beamed long notes
        0 /    5 duplicate ids
        0 /    1 empty containers
        0 /    5 missing parents
        0 /    4 notes on wrong clef
        0 /    4 out of range pitches
        0 /    2 overlapping text spanners
        0 /    2 unmatched stop text spans
        0 /    0 unterminated hairpins
        0 /    2 unterminated text spanners

    """
    violators, total = [], 0

    def key(wrapper):
        if isinstance(wrapper.unbundle_indicator(), _indicators.StartTextSpan):
            priority = 1
        else:
            priority = 0
        return (wrapper.leaked_start_offset, priority)

    name_to_wrappers = _aggregate_context_wrappers(argument)
    for name, wrappers in name_to_wrappers.items():
        wrappers.sort(key=key)
        open_spanners: dict = {}
        for wrapper in wrappers:
            if isinstance(wrapper.unbundle_indicator(), _indicators.StartTextSpan):
                total += 1
                command = wrapper.unbundle_indicator().command
                command = command.replace("start", "")
                command = command.replace("Start", "")
                if command not in open_spanners:
                    open_spanners[command] = []
                if open_spanners[command]:
                    violators.append(wrapper.component)
                open_spanners[command].append(wrapper.component)
            elif isinstance(wrapper.unbundle_indicator(), _indicators.StopTextSpan):
                command = wrapper.unbundle_indicator().command
                command = command.replace("stop", "")
                command = command.replace("Stop", "")
                if command in open_spanners and open_spanners[command]:
                    open_spanners[command].pop()
    return violators, total


def check_unmatched_stop_text_spans(argument) -> tuple[list, int]:
    r"""
    Checks unmatched stop text spans.

    ..  container:: example

        Unmatched stop text span is not wellformed:

        >>> voice = abjad.Voice("c'4 c'4 c'4 c'4")
        >>> stop_text_span = abjad.StopTextSpan()
        >>> abjad.attach(stop_text_span, voice[-1])
        >>> string = abjad.lilypond(voice)
        >>> print(string)
        \new Voice
        {
            c'4
            c'4
            c'4
            c'4
            \stopTextSpan
        }

        >>> count, string = abjad.wf.tabulate_wellformedness(voice)
        >>> print(string)
        0 /    4 beamed long notes
        0 /    5 duplicate ids
        0 /    1 empty containers
        0 /    5 missing parents
        0 /    4 notes on wrong clef
        0 /    4 out of range pitches
        0 /    0 overlapping text spanners
        1 /    0 unmatched stop text spans
        0 /    0 unterminated hairpins
        0 /    0 unterminated text spanners

        Matched stop text span is wellformed:

        >>> voice = abjad.Voice("c'4 c'4 c'4 c'4")
        >>> start_text_span = abjad.StartTextSpan()
        >>> abjad.attach(start_text_span, voice[0])
        >>> stop_text_span = abjad.StopTextSpan()
        >>> abjad.attach(stop_text_span, voice[-1])
        >>> abjad.show(voice) # doctest: +SKIP

        ..  docs::

            >>> string = abjad.lilypond(voice)
            >>> print(string)
            \new Voice
            {
                c'4
                \startTextSpan
                c'4
                c'4
                c'4
                \stopTextSpan
            }

        >>> abjad.wf.wellformed(voice)
        True

    """
    violators, total = [], 0
    name_to_wrappers = _aggregate_context_wrappers(argument)
    for name, wrappers in name_to_wrappers.items():
        wrappers.sort(key=lambda _: _.leaked_start_offset)
        open_spanners: dict = {}
        for wrapper in wrappers:
            if isinstance(wrapper.unbundle_indicator(), _indicators.StartTextSpan):
                total += 1
                command = wrapper.unbundle_indicator().command
                command = command.replace("start", "")
                command = command.replace("Start", "")
                if command not in open_spanners:
                    open_spanners[command] = []
                open_spanners[command].append(wrapper.component)
            elif isinstance(wrapper.unbundle_indicator(), _indicators.StopTextSpan):
                command = wrapper.unbundle_indicator().command
                command = command.replace("stop", "")
                command = command.replace("Stop", "")
                if command not in open_spanners or not open_spanners[command]:
                    violators.append(wrapper.component)
                else:
                    open_spanners[command].pop()
    return violators, total


def check_unterminated_hairpins(argument) -> tuple[list, int]:
    r"""
    Checks unterminated hairpins.

    ..  container:: example

        Unterminated crescendo is not wellformed:

        >>> voice = abjad.Voice("c'4 c'4 c'4 c'4")
        >>> start_hairpin = abjad.StartHairpin('<')
        >>> abjad.attach(start_hairpin, voice[0])
        >>> string = abjad.lilypond(voice)
        >>> print(string)
        \new Voice
        {
            c'4
            \<
            c'4
            c'4
            c'4
        }

        >>> count, string = abjad.wf.tabulate_wellformedness(voice)
        >>> print(string)
        0 /    4 beamed long notes
        0 /    5 duplicate ids
        0 /    1 empty containers
        0 /    5 missing parents
        0 /    4 notes on wrong clef
        0 /    4 out of range pitches
        0 /    0 overlapping text spanners
        0 /    0 unmatched stop text spans
        1 /    1 unterminated hairpins
        0 /    0 unterminated text spanners

        Even with start dynamic:

        >>> voice = abjad.Voice("c'4 c'4 c'4 c'4")
        >>> dynamic = abjad.Dynamic('f')
        >>> abjad.attach(dynamic, voice[0])
        >>> start_hairpin = abjad.StartHairpin('<')
        >>> abjad.attach(start_hairpin, voice[0])
        >>> string = abjad.lilypond(voice)
        >>> print(string)
        \new Voice
        {
            c'4
            \f
            \<
            c'4
            c'4
            c'4
        }

        >>> count, string = abjad.wf.tabulate_wellformedness(voice)
        >>> print(string)
        0 /    4 beamed long notes
        0 /    5 duplicate ids
        0 /    1 empty containers
        0 /    5 missing parents
        0 /    4 notes on wrong clef
        0 /    4 out of range pitches
        0 /    0 overlapping text spanners
        0 /    0 unmatched stop text spans
        1 /    1 unterminated hairpins
        0 /    0 unterminated text spanners

        Terminated crescendo is wellformed:

        >>> voice = abjad.Voice("c'4 c'4 c'4 c'4")
        >>> start_hairpin = abjad.StartHairpin('<')
        >>> abjad.attach(start_hairpin, voice[0])
        >>> dynamic = abjad.Dynamic('f')
        >>> abjad.attach(dynamic, voice[-1])
        >>> abjad.show(voice) # doctest: +SKIP

        ..  docs::

            >>> string = abjad.lilypond(voice)
            >>> print(string)
            \new Voice
            {
                c'4
                \<
                c'4
                c'4
                c'4
                \f
            }

        >>> abjad.wf.wellformed(voice)
        True

    ..  container:: example

        Bang-terminated crescendo is wellformed:

        >>> voice = abjad.Voice("c'4 c'4 c'4 c'4")
        >>> start_hairpin = abjad.StartHairpin('<')
        >>> abjad.attach(start_hairpin, voice[0])
        >>> stop_hairpin = abjad.StopHairpin()
        >>> abjad.attach(stop_hairpin, voice[-1])
        >>> abjad.show(voice) # doctest: +SKIP

        ..  docs::

            >>> string = abjad.lilypond(voice)
            >>> print(string)
            \new Voice
            {
                c'4
                \<
                c'4
                c'4
                c'4
                \!
            }

        >>> abjad.wf.wellformed(voice)
        True

    """
    violators, total = [], 0
    name_to_wrappers = _aggregate_context_wrappers(argument)
    for name, wrappers in name_to_wrappers.items():
        last_dynamic = None
        last_tag = _tag.Tag()
        wrappers.sort(key=lambda _: _.leaked_start_offset)
        for wrapper in wrappers:
            parameter = getattr(wrapper.unbundle_indicator(), "parameter", None)
            if parameter == "DYNAMIC" or isinstance(
                wrapper.unbundle_indicator(), _indicators.StopHairpin
            ):
                last_dynamic = wrapper.unbundle_indicator()
                last_tag = wrapper.tag
                if isinstance(wrapper.unbundle_indicator(), _indicators.StartHairpin):
                    total += 1
        if (
            isinstance(last_dynamic, _indicators.StartHairpin)
            and _tag.Tag("RIGHT_BROKEN").string not in last_tag.string
        ):
            violators.append(wrapper.component)
    return violators, total


def check_unterminated_text_spanners(argument) -> tuple[list, int]:
    r"""
    Checks unterminated text spanners.

    ..  container:: example

        Unterminated text spanner is not wellformed:

        >>> voice = abjad.Voice("c'4 c'4 c'4 c'4")
        >>> start_text_span = abjad.StartTextSpan()
        >>> abjad.attach(start_text_span, voice[0])
        >>> string = abjad.lilypond(voice)
        >>> print(string)
        \new Voice
        {
            c'4
            \startTextSpan
            c'4
            c'4
            c'4
        }

        >>> count, string = abjad.wf.tabulate_wellformedness(voice)
        >>> print(string)
        0 /    4 beamed long notes
        0 /    5 duplicate ids
        0 /    1 empty containers
        0 /    5 missing parents
        0 /    4 notes on wrong clef
        0 /    4 out of range pitches
        0 /    1 overlapping text spanners
        0 /    1 unmatched stop text spans
        0 /    0 unterminated hairpins
        1 /    1 unterminated text spanners

        Terminated text span is wellformed:

        >>> voice = abjad.Voice("c'4 c'4 c'4 c'4")
        >>> start_text_span = abjad.StartTextSpan()
        >>> abjad.attach(start_text_span, voice[0])
        >>> stop_text_span = abjad.StopTextSpan()
        >>> abjad.attach(stop_text_span, voice[-1])
        >>> abjad.show(voice) # doctest: +SKIP

        ..  docs::

            >>> string = abjad.lilypond(voice)
            >>> print(string)
            \new Voice
            {
                c'4
                \startTextSpan
                c'4
                c'4
                c'4
                \stopTextSpan
            }

        >>> abjad.wf.wellformed(voice)
        True

    """
    violators, total = [], 0
    name_to_wrappers = _aggregate_context_wrappers(argument)
    for name, wrappers in name_to_wrappers.items():
        wrappers.sort(key=lambda _: _.leaked_start_offset)
        open_spanners: dict = {}
        for wrapper in wrappers:
            if isinstance(wrapper.unbundle_indicator(), _indicators.StartTextSpan):
                total += 1
                command = wrapper.unbundle_indicator().command
                command = command.replace("start", "")
                command = command.replace("Start", "")
                if command not in open_spanners:
                    open_spanners[command] = []
                open_spanners[command].append(wrapper.component)
            elif isinstance(wrapper.unbundle_indicator(), _indicators.StopTextSpan):
                command = wrapper.unbundle_indicator().command
                command = command.replace("stop", "")
                command = command.replace("Stop", "")
                if command not in open_spanners or not open_spanners[command]:
                    # unmatched stop text span
                    pass
                else:
                    open_spanners[command].pop()
        for command, list_ in open_spanners.items():
            for component in list_:
                violators.append(component)
    return violators, total


_globals = globals()


def _call_functions(
    component,
    check_beamed_long_notes: bool = True,
    check_duplicate_ids: bool = True,
    check_empty_containers: bool = True,
    check_missing_parents: bool = True,
    check_notes_on_wrong_clef: bool = True,
    check_out_of_range_pitches: bool = True,
    check_overlapping_text_spanners: bool = True,
    check_unmatched_stop_text_spans: bool = True,
    check_unterminated_hairpins: bool = True,
    check_unterminated_text_spanners: bool = True,
):
    triples = []
    if check_beamed_long_notes:
        name = "check_beamed_long_notes"
        violators, count = _globals[name](component)
        triples.append((violators, count, name))
    if check_duplicate_ids:
        name = "check_duplicate_ids"
        violators, count = _globals[name](component)
        triples.append((violators, count, name))
    if check_empty_containers:
        name = "check_empty_containers"
        violators, count = _globals[name](component)
        triples.append((violators, count, name))
    if check_missing_parents:
        name = "check_missing_parents"
        violators, count = _globals[name](component)
        triples.append((violators, count, name))
    if check_notes_on_wrong_clef:
        name = "check_notes_on_wrong_clef"
        violators, count = _globals[name](component)
        triples.append((violators, count, name))
    if check_out_of_range_pitches:
        name = "check_out_of_range_pitches"
        violators, count = _globals[name](component)
        triples.append((violators, count, name))
    if check_overlapping_text_spanners:
        name = "check_overlapping_text_spanners"
        violators, count = _globals[name](component)
        triples.append((violators, count, name))
    if check_unmatched_stop_text_spans:
        name = "check_unmatched_stop_text_spans"
        violators, count = _globals[name](component)
        triples.append((violators, count, name))
    if check_unterminated_hairpins:
        name = "check_unterminated_hairpins"
        violators, count = _globals[name](component)
        triples.append((violators, count, name))
    if check_unterminated_text_spanners:
        name = "check_unterminated_text_spanners"
        violators, count = _globals[name](component)
        triples.append((violators, count, name))
    return triples


def tabulate_wellformedness(
    component,
    check_beamed_long_notes: bool = True,
    check_duplicate_ids: bool = True,
    check_empty_containers: bool = True,
    check_missing_parents: bool = True,
    check_notes_on_wrong_clef: bool = True,
    check_out_of_range_pitches: bool = True,
    check_overlapping_text_spanners: bool = True,
    check_unmatched_stop_text_spans: bool = True,
    check_unterminated_hairpins: bool = True,
    check_unterminated_text_spanners: bool = True,
) -> tuple[int, str]:
    """
    Tabulates wellformedness.
    """
    triples = _call_functions(
        component,
        check_beamed_long_notes=check_beamed_long_notes,
        check_duplicate_ids=check_duplicate_ids,
        check_empty_containers=check_empty_containers,
        check_missing_parents=check_missing_parents,
        check_notes_on_wrong_clef=check_notes_on_wrong_clef,
        check_out_of_range_pitches=check_out_of_range_pitches,
        check_overlapping_text_spanners=check_overlapping_text_spanners,
        check_unmatched_stop_text_spans=check_unmatched_stop_text_spans,
        check_unterminated_hairpins=check_unterminated_hairpins,
        check_unterminated_text_spanners=check_unterminated_text_spanners,
    )
    strings = []
    total_violators = 0
    for violators, total, name in triples:
        violator_count = len(violators)
        name = name.replace("check_", "")
        name = name.replace("_", " ")
        string = f"{violator_count} /    {total} {name}"
        strings.append(string)
        total_violators += violator_count
    return total_violators, "\n".join(strings)


def wellformed(
    component,
    check_beamed_long_notes: bool = True,
    check_duplicate_ids: bool = True,
    check_empty_containers: bool = True,
    check_missing_parents: bool = True,
    check_notes_on_wrong_clef: bool = True,
    check_out_of_range_pitches: bool = True,
    check_overlapping_text_spanners: bool = True,
    check_unmatched_stop_text_spans: bool = True,
    check_unterminated_hairpins: bool = True,
    check_unterminated_text_spanners: bool = True,
) -> bool:
    """
    Is true when ``component`` is wellformed.
    """
    triples = _call_functions(
        component,
        check_beamed_long_notes=check_beamed_long_notes,
        check_duplicate_ids=check_duplicate_ids,
        check_empty_containers=check_empty_containers,
        check_missing_parents=check_missing_parents,
        check_notes_on_wrong_clef=check_notes_on_wrong_clef,
        check_out_of_range_pitches=check_out_of_range_pitches,
        check_overlapping_text_spanners=check_overlapping_text_spanners,
        check_unmatched_stop_text_spans=check_unmatched_stop_text_spans,
        check_unterminated_hairpins=check_unterminated_hairpins,
        check_unterminated_text_spanners=check_unterminated_text_spanners,
    )
    for violators, total, name in triples:
        if violators:
            return False
    return True
